from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from pathlib import Path


option_keys = {
    "UnicodeKeyboard": "dword",
    "ChartAutoActivateCodepage": "dword",
    "UnicodeRangePercent": "dword",
    "AutoAlignVector": "dword",
    "ContourSnapAllPoints": "dword",
    "ShowMeterPanel": "dword",
    "FontLabNodes": "dword",
    "EditDeleteAlt": "dword",
    "ControlSwitch": "dword",
    "RemoveZeroKerning": "dword",
    "EditAutoSelectLayers": "dword",
    "SyncronizeMasters": "dword",
    "EditComponentsByOutline": "dword",
    "EditLayersExpanded": "dword",
    "EditKeyboardEditBCP": "dword",
    "EditShowMasterPoints": "dword",
    "EditConnectMasters": "dword",
    "EditShowMeasurementLine": "dword",
    "EditUndoSelection": "dword",
    "EditShowNodesSelection": "dword",
    "MaskMetrics": "dword",
    "EditEditSelection": "dword",
    "EditDoubleClickBitmap": "dword",
    "EditDoubleClickMask": "dword",
    "EditBCPsFixed": "dword",
    "EditPreviewKey": "dword",
    "OTPreviewSize": "dword",
    "OTSampleSize": "dword",
    "EditNotFillOpen": "dword",
    "OptimizeAlign": "dword",
    "OptimizeReduce": "dword",
    "UnicodeStrings": "dword",
    "DuplicateX": "dword",
    "DuplicateY": "dword",
    "PasteX": "dword",
    "PasteY": "dword",
    "Shift10": "dword",
    "MetricsBottomBar0": "dword",
    "MetricsShowTable0": "dword",
    "MetricsShowBar0": "dword",
    "MetricsBarExpanded0": "dword",
    "MetricsBottomBar1": "dword",
    "MetricsShowTable1": "dword",
    "MetricsShowBar1": "dword",
    "MetricsBarExpanded1": "dword",
    "MetricsBottomBar2": "dword",
    "MetricsShowTable2": "dword",
    "MetricsShowBar2": "dword",
    "MetricsBarExpanded2": "dword",
    "MetricsBottomBar3": "dword",
    "MetricsShowTable3": "dword",
    "MetricsShowBar3": "dword",
    "MetricsBarExpanded3": "dword",
    "FontExpanded": "dword",
    "AutoRemoveHints": "dword",
    "OTAddClasses": "dword",
    "OTAddMetricsClasses": "dword",
    "OTWriteGDEF": "dword",
    "OTWriteMort": "dword",
    "OTReadMort": "dword",
    "OTGenerate": "dword",
    "OTCompileFeatures": "dword",
    "OTOldContextRule": "dword",
    "TTIReadKernFeature": "dword",
    "TTIGenerateNames": "dword",
    "TTEWriteKernFeature": "dword",
    "TTEWriteKernTable": "dword",
    "TTESmartMacNames": "dword",
    "TTESmartFontName": "dword",
    "TTEVersionOS2": "dword",
    "TTEExportOT": "dword",
    "TTEExportVOLT": "dword",
    "TTIReadOT": "dword",
    "TTIReadBinaryOT": "dword",
    "TTIStoreBinaryOT": "dword",
    "TTToolReverseShift": "dword",
    "TTEWriteBitmaps": "dword",
    "TTIReadBitmaps": "dword",
    "CopyHDMXData": "dword",
    "TracerCurves": "dword",
    "TracerExtremePoints": "dword",
    "TracerTolerance": "dword",
    "TracerStraightenAngle": "dword",
    "TracerPresetMode": "dword",
    "TracerCurveFit": "dword",
    "TTIFontNames": "dword",
    "TTEFontNames": "dword",
    "BitmapSize": "dword",
    "ColorBackground": "dword",
    "ColorBackgroundMask": "dword",
    "ColorBackgroundMetrics": "dword",
    "ColorForegroundMetrics": "dword",
    "ColorMaskPen": "dword",
    "ColorClosepath": "dword",
    "ColorSeacPen": "dword",
    "ColorHintsPen": "dword",
    "ColorTemplate": "dword",
    "ColorOutline": "dword",
    "ColorBitmap": "dword",
    "ColorVMetrics": "dword",
    "ColorMetrics": "dword",
    "ColorEcho": "dword",
    "ColorGrid": "dword",
    "ColorGuide": "dword",
    "ColorGlobalGuide": "dword",
    "ColorGroups": "dword",
    "ColorNeighbors": "dword",
    "StyleInactiveMetrics": "dword",
    "StyleMetrics": "dword",
    "SamplePPM1": "dword",
    "SamplePPM2": "dword",
    "HideAllLayers": "dword",
    "CustomMarkColor": "dword",
    "QuickZoomScale": "dword",
    "TrackingMode": "dword",
    "TrackingDistance": "dword",
    "FitAscender": "dword",
    "FitDescender": "dword",
    "EraseSize": "dword",
    "ClosepathArrowLen": "dword",
    "DefaultGlyph": "dword",
    "GlyphNameSortMode": "dword",
    "EditBitmapStyle": "dword",
    "EditZoomMode": "dword",
    "EditHitDistance": "dword",
    "EditChangeCursor": "dword",
    "EditHandleTool": "dword",
    "EditBCVOpacity": "dword",
    "EditFollowScroll": "dword",
    "EditLeaveEcho": "dword",
    "EditInstantRefresh": "dword",
    "EditShowSelection": "dword",
    "EditShowAnchorNames": "dword",
    "EditSmoothOutline": "dword",
    "EditAllSmooth": "dword",
    "EditSmallNodes": "dword",
    "EditThickOutline": "dword",
    "EditShowCross": "dword",
    "EditScaleEPS": "dword",
    "MMExtrapolation": "dword",
    "EditShowPosition": "dword",
    "HideToolbars": "dword",
    "AutoMetricsLeft": "dword",
    "AutoMetricsRight": "dword",
    "AutoMetricsClose": "dword",
    "AutohintingMinHLen": "dword",
    "AutohintingMinVLen": "dword",
    "AutohintingMinHWidth": "dword",
    "AutohintingMinVWidth": "dword",
    "AutohintingMaxHWidth": "dword",
    "AutohintingMaxVWidth": "dword",
    "AutohintingRemoveHints": "dword",
    "AutohintingEngine": "dword",
    "CreateUnexistingCharacters": "dword",
    "AutohintingHRatio": "str",
    "AutohintingVRatio": "str",
    "PaintTextFont": "str",
    "PreviewPointSize": "str",
    "PreviewPPMs": "str",
    "PreviewPPMsExpanded": "str",
    "PreviewExpandedHeight": "dword",
    "PreviewWidth": "dword",
    "PreviewSmooth": "dword",
    "PreviewExpanded": "dword",
    "PreviewBlueColor": "dword",
    "PreviewHighlightKeyGlyphs": "dword",
    "PreviewLockDependences": "dword",
    "PreviewKeepExceptions": "dword",
    "PreviewShowReferences": "dword",
    "PreviewApplyTemplate": "dword",
    "PreviewAutoWrap": "dword",
    "PreviewKerningIcons": "dword",
    "PreviewFocusString": "dword",
    "PreviewPanelTop": "dword",
    "PreviewMode": "dword",
    "PreviewSize": "dword",
    "PreviewSecondLine": "str",
    "VendorCode": "str",
    "ExpandKernFlags": "dword",
    "ExpandKernCount": "dword",
    "ExpandKernCodepage": "str",
    "CacheTTPath": "str",
    "CacheTTUse": "dword",
    "Template_Use": "dword",
    "Template_Path": "str",
    "Template_BDF": "str",
    "DSIG_Use": "dword",
    "DSIG_TimeStamp": "dword",
    "DSIG_SertPath": "str",
    "DSIG_KeyPath": "str",
    "DSIG_Password": "str",
    "FileOpenSample": "str",
    "PreviewAlternativeArrows": "dword",
    "PreviewAutoScroll": "dword",
    "PreviewListWidth": "dword",
    "PreviewFilterKerning": "dword",
    "PreviewFilterMetrics": "dword",
    "PreviewPaintStyle": "dword",
    "PreviewLockStyle": "dword",
    "PreviewSnapStyle": "dword",
    "OverlapMode": "dword",
    "PaintStyleEx": "dword",
    "SnapStyleEx": "dword",
    "LockStyleEx": "dword",
    "EditGridX": "dword",
    "EditGridY": "dword",
    "EditRulers": "dword",
    "EditSmoothIsSmooth": "dword",
    "PaintColor": "dword",
    "PaintTool": "dword",
    "PaintBrushWidth": "dword",
    "PaintBrushRoundness": "dword",
    "PaintBrushBody": "dword",
    "PaintBrushVectorX": "dword",
    "PaintBrushVectorY": "dword",
    "PaintMode": "dword",
    "PaintBrushJoin": "dword",
    "PaintBrushCap": "dword",
    "PaintTracePolygon": "dword",
    "PaintTextSize": "dword",
    "PaintAutoTransform": "dword",
    "PaintAutoView": "dword",
    "PaintReverseBrushAutoColor": "dword",
    "PaintJapanese": "dword",
    "FontAudit": "dword",
    "FontAuditLive": "dword",
    "TTIAutohint": "dword",
    "TTIDecompose": "dword",
    "TTIKeepHints": "dword",
    "TTIConvert": "dword",
    "TTIScale1000": "dword",
    "TTIStoreTables": "dword",
    "TTEStoreTables": "dword",
    "TTECmap10": "dword",
    "EmbeddingType": "dword",
    "TTHHintingOptions": "dword",
    "TTHPPM": "dword",
    "ATMPPM": "dword",
    "TTEExportUnicode": "dword",
    "TTEHint": "dword",
    "TTEAutohint": "dword",
    "TTEKeep": "dword",
    "TTEVisual": "dword",
    "TTEheadBBoxSavings": "dword",
    "TTEApplyBBoxSavings": "dword",
    "TTEAutoWinAscDesc": "dword",
    "PreviewProcessFeatures": "dword",
    "PreviewShowClassMembers": "dword",
    "TTEAddCharacters": "dword",
    "TTENoReorder": "dword",
    "TTESubrize": "dword",
    "TTEDecompose": "dword",
    "T1Terminal": "dword",
    "AutoSave": "dword",
    "AutoSaveTick": "dword",
    "ChartShowNotes": "dword",
    "ChartSorting": "dword",
    "ChartDragDrop": "dword",
    "GroupKerning": "dword",
    "GroupFill": "dword",
    "GroupShapeFill": "dword",
    "GroupMask": "dword",
    "GroupShapeMask": "dword",
    "GroupShapeOpacity": "dword",
    "GroupUseMetrics": "dword",
    "GroupCenter": "dword",
    "GroupDoubleClick": "dword",
    "GroupShiftX": "dword",
    "GroupShiftY": "dword",
    "T1Sort": "dword",
    "T1UseOTFamilyName": "dword",
    "T1UseOTStyleName": "dword",
    "T1Encoding": "dword",
    "T1Decompose": "dword",
    "T1MatchEncoding": "dword",
    "T1PFM": "dword",
    "T1AFM": "dword",
    "T1FSType": "dword",
    "T1Autohint": "dword",
    "T1Unicode": "dword",
    "T1UseOS2": "dword",
    "T1UseTrademarkName": "dword",
    "T1ExportEncoding": "dword",
    "CodepagesCount": "dword",
    "OpenOutput": "dword",
    "Backup": "dword",
    "ChartUseTemplate": "dword",
    "ChartCopyKerning": "dword",
    "GlyphsBarOpen": "dword",
    "GlyphsBarTop": "dword",
    "ChartDoubleClick": "dword",
    "ChartAutoHide": "dword",
    "ChartPreviewMarks": "dword",
    "ChartSampleSize": "dword",
    "ChartStyle": "dword",
    "ChartCaptionFontSize": "dword",
    "ChartTemplateFontSize": "dword",
    "ChartCaptionFontCharSet": "dword",
    "ChartTemplateFontCharSet": "dword",
    "ChartGenerate": "dword",
    "ChartGenerateRTL": "dword",
    "ChartAutoUnicode": "dword",
    "ChartApplyTemplate": "dword",
    "ChartCaptionFontName": "str",
    "ChartTemplateFontName": "str",
    "FontSmallName": "str",
    "FontSmallSize": "dword",
    "FontDialogName": "str",
    "FontDialogSize": "dword",
    "FontRulerName": "str",
    "FontRulerSize": "dword",
    "FontPanelName": "str",
    "FontPanelSize": "dword",
    "FontTableName": "str",
    "FontTableSize": "dword",
    "FontFixedName": "str",
    "FontFixedSize": "dword",
    "T1ExportCompatibleCyrillic": "dword",
    "OutputFontName": "str",
    "MacroFontName": "str",
    "CustomDict": "str",
    "EnablePython": "dword",
    "EnableExtPythonEditor": "dword",
    "ExtPythonEditor": "str",
    "EnableTooltips": "dword",
    "EnableStickPanels": "dword",
    "ATMScaler": "dword",
    "ATMSmooth": "dword",
    "TTFScaler": "dword",
    "TTFSmooth": "dword",
}


def parse_registry_file(file_path: Path) -> Dict[str, Dict[str, str | int | float]]:
    with open(file_path, "r") as f:
        data = f.read()

    reg = {}
    lines = data.splitlines()
    if not lines:
        raise EOFError

    assert lines[0] == "Windows Registry Editor Version 5.00"

    key = None
    values = {}

    for line in lines[1:]:
        if line == "":
            continue

        if line.startswith("[") and line.endswith("]"):
            if key:
                reg[key] = values
            key = line.strip("[]")
            values = {}

        elif line.startswith('"') and '"=' in line:
            k, v = line.split("=", 1)
            values[k.strip('"')] = v

        else:
            print("Unhandled line in registry file:")
            print(line)
            raise ValueError

    if key:
        reg[key] = values

    # We now have a dict where the keys may be unknown, and the values are in raw format

    parsed = {}

    reg_options = reg.get(r"HKEY_CURRENT_USER\Software\FontLab\FontStudio 5\Options")
    if reg_options is None:
        print("Could not find data section in registry file")
        raise ValueError

    for key, val_type in option_keys.items():
        if key not in reg_options:
            continue

        if val_type == "dword":
            hex_str = reg_options[key].split(":", 1)[1]
            assert len(hex_str) == 8
            val = int(hex_str, 16)

        elif val_type == "str":
            s = reg_options[key]
            assert s.startswith('"')
            assert s.endswith('"')
            val = reg_options[key][1:-1]

        else:
            print(f"Unhandled registry value type {val_type}")
            raise ValueError

        parsed[key] = val

    return parsed



import unittest
from pathlib import Path

import pytest

from FL import Feature, Font, Glyph, fl


class FontTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        f = Font()
        assert isinstance(f, Font)

    def test_instantiation_path(self) -> None:
        vfb_path = Path(__file__).parent.parent.parent / "tests" / "data" / "empty.vfb"

        f = Font(str(vfb_path))
        assert len(fl) == 0
        fl.Add(f)
        assert len(fl) == 1
        assert f.file_name == str(vfb_path)

    def test_add_glyph(self) -> None:
        f = Font()
        assert len(f) == 0
        f.glyphs.append(Glyph())
        assert len(f) == 1

    def test_find_glyph(self) -> None:
        f = Font()
        g = Glyph()
        g.name = "adieresis"
        f.glyphs.append(g)

        # Glyph in font with index 0
        i = f.FindGlyph("adieresis")
        assert i == 0

        # Glyph not in font
        i = f.FindGlyph("a")
        assert i == -1

    def test_features_readonly(self) -> None:
        f = Font()
        with pytest.raises(RuntimeError):
            f.features = []

    def test_features_iadd_freeze(self) -> None:
        f = Font()
        with pytest.raises(ReferenceError):
            f.features += Feature()

    def test_features_setitem(self) -> None:
        f = Font()
        fea = Feature("aaaa")
        f.features.append(fea)
        assert len(f.features) == 1
        assert fea.parent == f
        fea = Feature("bbbb")
        f.features[0] = fea
        assert len(f.features) == 1
        assert fea.parent == f
        assert f.features[0].tag == "bbbb"

    def test_features_setslice(self) -> None:
        # FIXME
        f = Font()
        f.features.append(Feature("aaaa"))
        f.features.append(Feature("bbbb"))
        assert len(f.features) == 2
        # Will actually raise an IndexError, only [0:1] works in FL if the feature
        # list has two elements:
        # f.features[0:2] = [Feature("cccc"), Feature("dddd")]
        # assert f.features[0].parent == f
        # assert f.features[0].tag == "cccc"
        # assert f.features[1].parent == f
        # assert f.features[1].tag == "dddd"

    def test_features_append(self) -> None:
        f = Font()
        fea = Feature()
        f.features.append(fea)
        assert len(f.features) == 1
        assert fea.parent == f

    def test_fontnames_readonly(self) -> None:
        f = Font()
        with pytest.raises(RuntimeError):
            f.fontnames = []

    def test_glyphs_readonly(self) -> None:
        f = Font()
        with pytest.raises(RuntimeError):
            f.glyphs = []

    def test_truetypetables_readonly(self) -> None:
        f = Font()
        with pytest.raises(RuntimeError):
            f.truetypetables = []

    def test_write_empty_from_scratch(self) -> None:
        f = Font()
        vfb_path = Path(__file__).parent.parent / "data" / "empty.scratch.vfb"
        f.Save(str(vfb_path))

    def test_SaveAFM(self) -> None:
        base_path = Path(__file__).parent.parent / "data" / "mini.vfb"
        f = Font(str(base_path))
        f.SaveAFM(str(base_path.with_suffix(".gen.afm")))
        with open(base_path.with_suffix(".gen.afm")) as afm:
            actual = afm.read()
        with open(base_path.with_suffix(".afm")) as afm:
            expected = afm.read()
        assert actual == expected

    def test_SaveAFM_expanded(self) -> None:
        base_path = Path(__file__).parent.parent / "data" / "mini.vfb"
        f = Font(str(base_path))
        assert f.classes == ["_LAT_a_LEFT: a' c", "_LAT_c_RIGHT: c' a", "_LAT_b: b'"]
        kerning = f.fake_get_afm_kerning(expand_kerning=True)
        assert kerning == [
            ("a", "period", -100),
            ("c", "period", -100),  # expanded pair
            ("period", "a", -90),
            ("period", "c", -120),
            ("period", "period", 10),
            ("space", "period", -80),
        ]
        f.fake_save_afm_expanded(str(base_path.with_suffix(".expanded.gen.afm")))
        with open(base_path.with_suffix(".expanded.gen.afm")) as afm:
            actual = afm.read()
        with open(base_path.with_suffix(".expanded.afm")) as afm:
            expected = afm.read()
        assert actual == expected

    def test_bounding_box(self) -> None:
        base_path = Path(__file__).parent.parent / "data" / "mini.vfb"
        f = Font(str(base_path))
        bbox = f.fake_bounding_rect()
        assert (bbox.ll.x, bbox.ll.y, bbox.ur.x, bbox.ur.y) == (66, -66, 531, 646)

    def test_classes(self) -> None:
        f = Font()
        g = Glyph()
        g.name = "adieresis"
        f.glyphs.append(g)

        # Glyph in font with index 0
        i = f.FindGlyph("adieresis")
        assert i == 0

        # Glyph not in font
        i = f.FindGlyph("a")
        assert i == -1

    def test_classes_deserialize(self) -> None:
        base_path = Path(__file__).parent.parent / "data" / "classes_duplicate.vfb"
        f = Font(str(base_path))
        assert f.classes == [
            "_a: a'",
            "_a: a'",
            ".mtrx3: a'",
            ".mtrx4:",
            ".mtrx5:",
            "class6:",
            "_kern7:",
            ".mtrx8:",
        ]
        assert f._classes._kerning_flags == {"_a": (1024, 0), "_kern7": (2048, 0)}
        assert f._classes._metrics_flags == {
            ".mtrx3": (
                0,
                1025,
                0,
            ),
            ".mtrx4": (
                0,
                5121,
                0,
            ),
            ".mtrx5": (
                0,
                6145,
                0,
            ),
            ".mtrx8": (
                0,
                3073,
                0,
            ),
        }
        # First class is kerning left
        assert f.GetClassLeft(0) == 1
        assert f.GetClassRight(0) == 0
        assert f.GetClassMetricsFlags(0) == (0, 0, 0)
        # Duplicate class has no flags
        assert f.GetClassLeft(1) == 0
        assert f.GetClassRight(1) == 0
        assert f.GetClassMetricsFlags(0) == (0, 0, 0)
        # Class 2 is metrics left
        assert f.GetClassLeft(2) == 0
        assert f.GetClassRight(2) == 0
        assert f.GetClassMetricsFlags(2) == (1, 0, 0)
        # Class 3 is metrics left and width
        assert f.GetClassLeft(3) == 0
        assert f.GetClassRight(3) == 0
        assert f.GetClassMetricsFlags(3) == (1, 0, 1)
        # Class 4 is metrics right and width
        assert f.GetClassLeft(4) == 0
        assert f.GetClassRight(4) == 0
        assert f.GetClassMetricsFlags(4) == (0, 1, 1)
        # Class 5 is OpenType
        assert f.GetClassLeft(5) == 0
        assert f.GetClassRight(5) == 0
        assert f.GetClassMetricsFlags(5) == (0, 0, 0)
        # Class 6 is kerning right
        assert f.GetClassLeft(6) == 0
        assert f.GetClassRight(6) == 1
        assert f.GetClassMetricsFlags(6) == (0, 0, 0)
        # Class 7 is metrics left and right
        assert f.GetClassLeft(7) == 0
        assert f.GetClassRight(7) == 0
        assert f.GetClassMetricsFlags(7) == (1, 1, 0)
        # Class 8 is out of bounds
        assert f.GetClassLeft(8) is None
        assert f.GetClassRight(8) is None
        assert f.GetClassMetricsFlags(8) is None

    def test_class_flags_persist(self) -> None:
        # Class flags should persist in the font when the class list is replaced
        f = Font()
        f.classes = ["caps: O", ".mtrx1: O", "_kern3: O'"]
        f.SetClassFlags(2, 1, 0)  # 3rd class is left
        assert (f.GetClassLeft(0), f.GetClassRight(0)) == (0, 0)
        assert (f.GetClassLeft(1), f.GetClassRight(1)) == (0, 0)
        assert (f.GetClassLeft(2), f.GetClassRight(2)) == (1, 0)
        assert (f.GetClassLeft(3), f.GetClassRight(3)) == (None, None)

        f.classes = ["caps: O", "_kern3: O'", ".mtrx1: O", "_A: A' Aacute"]
        # Class "_kern3" should still have the left flag, but it is #1 now
        assert (f.GetClassLeft(1), f.GetClassRight(1)) == (1, 0)
        assert (f.GetClassLeft(3), f.GetClassRight(3)) == (0, 0)
        assert (f.GetClassLeft(4), f.GetClassRight(4)) == (None, None)

        f.classes = ["caps: O", ".mtrx1: O", "_kern3: O'"]
        # 3rd class should still have the left flag
        assert (f.GetClassLeft(2), f.GetClassRight(2)) == (1, 0)
        assert (f.GetClassLeft(3), f.GetClassRight(3)) == (None, None)

    def test_encoding_deserialize(self) -> None:
        base_path = Path(__file__).parent.parent / "data" / "mini.vfb"
        f = Font(str(base_path))
        encoding = [e.name for e in f.encoding]
        assert encoding == [
            "_0000",
            "_0001",
            "_0002",
            "_0003",
            "_0004",
            "breve",
            "dotaccent",
            "_0007",
            "ring",
            "_0009",
            "hungarumlaut",
            "ogonek",
            "caron",
            "dotlessi",
            "_0014",
            "_0015",
            "_0016",
            "_0017",
            "_0018",
            "_0019",
            "_0020",
            "_0021",
            "_0022",
            "_0023",
            "_0024",
            "fraction",
            "fi",
            "fl",
            "Lslash",
            "lslash",
            "Zcaron",
            "zcaron",
            "space",
            "exclam",
            "quotedbl",
            "numbersign",
            "dollar",
            "percent",
            "ampersand",
            "quotesingle",
            "parenleft",
            "parenright",
            "asterisk",
            "plus",
            "comma",
            "hyphen",
            "period",
            "slash",
            "zero",
            "one",
            "two",
            "three",
            "four",
            "five",
            "six",
            "seven",
            "eight",
            "nine",
            "colon",
            "semicolon",
            "less",
            "equal",
            "greater",
            "question",
            "at",
            "A",
            "B",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "J",
            "K",
            "L",
            "M",
            "N",
            "O",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "U",
            "V",
            "W",
            "X",
            "Y",
            "Z",
            "bracketleft",
            "backslash",
            "bracketright",
            "asciicircum",
            "underscore",
            "grave",
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "j",
            "k",
            "l",
            "m",
            "n",
            "o",
            "p",
            "q",
            "r",
            "s",
            "t",
            "u",
            "v",
            "w",
            "x",
            "y",
            "z",
            "braceleft",
            "bar",
            "braceright",
            "asciitilde",
            "_0127",
            "Euro",
            "_0129",
            "quotesinglbase",
            "florin",
            "quotedblbase",
            "ellipsis",
            "dagger",
            "daggerdbl",
            "circumflex",
            "perthousand",
            "Scaron",
            "guilsinglleft",
            "OE",
            "_0141",
            "_0142",
            "_0143",
            "_0144",
            "quoteleft",
            "quoteright",
            "quotedblleft",
            "quotedblright",
            "bullet",
            "endash",
            "emdash",
            "tilde",
            "trademark",
            "scaron",
            "guilsinglright",
            "oe",
            "_0157",
            "_0158",
            "Ydieresis",
            "uni00A0",
            "exclamdown",
            "cent",
            "sterling",
            "currency",
            "yen",
            "brokenbar",
            "section",
            "dieresis",
            "copyright",
            "ordfeminine",
            "guillemotleft",
            "logicalnot",
            "minus",
            "registered",
            "macron",
            "degree",
            "plusminus",
            "twosuperior",
            "threesuperior",
            "acute",
            "mu",
            "paragraph",
            "periodcentered",
            "cedilla",
            "onesuperior",
            "ordmasculine",
            "guillemotright",
            "onequarter",
            "onehalf",
            "threequarters",
            "questiondown",
            "Agrave",
            "Aacute",
            "Acircumflex",
            "Atilde",
            "Adieresis",
            "Aring",
            "AE",
            "Ccedilla",
            "Egrave",
            "Eacute",
            "Ecircumflex",
            "Edieresis",
            "Igrave",
            "Iacute",
            "Icircumflex",
            "Idieresis",
            "Eth",
            "Ntilde",
            "Ograve",
            "Oacute",
            "Ocircumflex",
            "Otilde",
            "Odieresis",
            "multiply",
            "Oslash",
            "Ugrave",
            "Uacute",
            "Ucircumflex",
            "Udieresis",
            "Yacute",
            "Thorn",
            "germandbls",
            "agrave",
            "aacute",
            "acircumflex",
            "atilde",
            "adieresis",
            "aring",
            "ae",
            "ccedilla",
            "egrave",
            "eacute",
            "ecircumflex",
            "edieresis",
            "igrave",
            "iacute",
            "icircumflex",
            "idieresis",
            "eth",
            "ntilde",
            "ograve",
            "oacute",
            "ocircumflex",
            "otilde",
            "odieresis",
            "divide",
            "oslash",
            "ugrave",
            "uacute",
            "ucircumflex",
            "udieresis",
            "yacute",
            "thorn",
            "ydieresis",
        ]



from fakeLabDemo.selection.composites import selectComposites

from FL import Component, Font, Glyph, Point, fl


def test_selectComposites():
    # Construct a fake FontLab font object
    font = Font()
    g = Glyph(1)
    g.name = "A"
    g.width = 500
    g.unicode = 0x41
    font.glyphs.append(g)

    g = Glyph(1)
    g.name = "dieresis"
    g.width = 500
    g.unicode = 0xA8
    font.glyphs.append(g)

    g = Glyph(1)
    g.name = "Adieresis"
    g.width = 500
    g.unicode = 0xC4
    g.components.append(Component(0))
    g.components.append(Component(1, Point(0, 300)))
    font.glyphs.append(g)

    # Add the font to the FL object
    fl.Add(font)
    fl.UpdateFont()

    # Run our script to be tested on the font
    selectComposites(fl.font)

    # You could save the fake font to JSON instead of VFB.
    # fl.font.Save("test_composites.vfb.json")

    # Test if the correct glyphs have been selected
    assert fl.Selected(0) == 0
    assert fl.Selected(1) == 0
    assert fl.Selected(2) == 1

    # Close the fake font
    fl.Close()



import unittest

from FL.objects.Matrix import Matrix
from FL.objects.Point import Point


class MatrixTests(unittest.TestCase):
    def test_instantiation(self):
        m = Matrix()
        assert m.a == 1
        assert m.b == 0
        assert m.c == 0
        assert m.d == 1
        assert m.e == 0
        assert m.f == 0

    # Addition

    def test_add(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        mr = m1 + m2
        assert mr.a == 5
        assert mr.b == 9
        assert mr.c == 15
        assert mr.d == 25
        assert mr.e == 43
        assert mr.f == 77

    def test_iadd(self):
        m = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        m += m2
        assert m.a == 5
        assert m.b == 9
        assert m.c == 15
        assert m.d == 25
        assert m.e == 43
        assert m.f == 77

    def test_Add(self):
        m = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        m.Add(m2)
        assert m.a == 5
        assert m.b == 9
        assert m.c == 15
        assert m.d == 25
        assert m.e == 43
        assert m.f == 77

    # Subtraction

    def test_sub(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        mr = m1 - m2
        assert mr.a == -1
        assert mr.b == -1
        assert mr.c == 1
        assert mr.d == 7
        assert mr.e == 21
        assert mr.f == 51

    def test_isub(self):
        m = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        m -= m2
        assert m.a == -1
        assert m.b == -1
        assert m.c == 1
        assert m.d == 7
        assert m.e == 21
        assert m.f == 51

    def test_Sub(self):
        m = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        m.Sub(m2)
        assert m.a == -1
        assert m.b == -1
        assert m.c == 1
        assert m.d == 7
        assert m.e == 21
        assert m.f == 51

    # Multiplication

    def test_mul_float(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m1 = m1 * 3
        assert m1.a == 6
        assert m1.b == 6
        assert m1.c == 6
        assert m1.d == 6
        assert m1.e == 6
        assert m1.f == 6

    def test_mul_matrix(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m2 = Matrix(3, 3, 3, 3, 3, 3)
        m1 = m1 * m2
        assert m1.a == 12
        assert m1.b == 12
        assert m1.c == 12
        assert m1.d == 12
        assert m1.e == 15
        assert m1.f == 15

    def test_imul_float(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m1 *= 3
        assert m1.a == 6
        assert m1.b == 6
        assert m1.c == 6
        assert m1.d == 6
        assert m1.e == 6
        assert m1.f == 6

    def test_imul_matrix(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m2 = Matrix(3, 3, 3, 3, 3, 3)
        m1 *= m2
        assert m1.a == 12
        assert m1.b == 12
        assert m1.c == 12
        assert m1.d == 12
        assert m1.e == 15
        assert m1.f == 15

    def test_Mul_float(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m1.Mul(3)
        assert m1.a == 6
        assert m1.b == 6
        assert m1.c == 6
        assert m1.d == 6
        assert m1.e == 6
        assert m1.f == 6

    # Transform

    def test_Transform(self):
        m1 = Matrix(2, 2, 2, 2, 2, 2)
        m2 = Matrix(3, 3, 3, 3, 3, 3)
        m1 = m1 * m2
        assert m1.a == 12
        assert m1.b == 12
        assert m1.c == 12
        assert m1.d == 12
        assert m1.e == 15
        assert m1.f == 15

    def test_Transform_2(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        m2 = Matrix(3, 5, 7, 9, 11, 13)
        m1 = m1 * m2
        assert m1.a == 46
        assert m1.b == 92
        assert m1.c == 86
        assert m1.d == 172
        assert m1.e == 427
        assert m1.f == 813

    # FakeLab addition

    def test_Transform_Point(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        p = Point(44, 55)
        p.Transform(m1)
        assert (p.x, p.y) == (340, 1296)

    def test_Transform_Point_float(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        p = Point(44.2, 55.5)
        p.Transform(m1)
        assert (p.x, p.y) == (342.4, 1305.6)



from __future__ import annotations

from math import atan2, degrees, radians, tan
from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph
    from FL.objects.Matrix import Matrix


class Guide(Copyable):
    # Constructor

    __slots__ = ["_angles", "_color", "_name", "_parent", "_positions", "_widths"]

    def __init__(
        self, guide_or_position: Guide | int | None = None, angle: float = 0.0
    ) -> None:
        """
        Guide - class to represent guideline

        Guide()                - generic constructor, creates a Guide with zero coordinates
        Guide(Guide)           - copy constructor
        Guide(position)        - creates a Guide and assigns position
        Guide(position, angle) - creates a Guide and assigns position and width values

        Args:
            guide_or_position (Guide | int | None, optional): The guide to be copied,
                or the position of the guide. Defaults to None.
            angle (float, optional): The width. Defaults to 0.0.
        """
        self._parent: Glyph | None = None
        self._positions: list[int] = [0] * 16
        self._widths: list[int] = [21] * 16  # WTF

        # Without API:
        self._color: str | None = None
        self._name: str | None = None

        # Process params

        if isinstance(guide_or_position, Guide):
            self._copy_constructor(guide_or_position)

        elif isinstance(guide_or_position, int):
            self.position = guide_or_position
            if angle is not None:
                self.angle = angle
        # else: Empty guide

    def __repr__(self) -> str:
        return f"<Guide pos: {self.position}, angle: {self.angle}>"

    @staticmethod
    def fake_angle_to_width(angle: float) -> int:
        return min(10000, round(tan(radians(angle)) * 10000))

    @staticmethod
    def fake_width_to_angle(width: float) -> float:
        return degrees(atan2(width, 10000))

    @property
    def angle(self) -> float:
        """The angle of the guideline.

        Returns:
            float: The angle of the guideline in degrees for the first master
        """
        return self.fake_width_to_angle(self.width)

    @angle.setter
    def angle(self, value: float) -> None:
        self.width = self.fake_angle_to_width(value)

    @property
    def parent(self) -> Glyph | None:
        """Guide's parent object, `Glyph`. If the guide is global, parent is an orphan
        `Glyph`.

        Returns:
            Glyph | None: The parent object.
        """
        return self._parent

    @property
    def position(self) -> int:
        """The position of the guideline. Setting the position sets the same value for
        all masters.

        Returns:
            int: Return position of the guideline for the first master
        """
        return self._positions[0]

    @position.setter
    def position(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError

        num_masters = len(self._positions)
        self._positions.clear()
        self._positions.extend([value] * num_masters)

    @property
    def positions(self) -> list[int]:
        return self._positions

    @positions.setter
    def positions(self, value: list[int]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "positions" of class Guide'
        )

    @property
    def width(self) -> float:
        """The "width" of the guideline. Setting the width sets the same value for
        all masters. Actually, the width is a representation of the angle:
        `width = round(tan(radians(value)) * 10000)`

        Returns:
            int: Return "width" of the guideline for the first master
        """
        return float(self._widths[0])

    @width.setter
    def width(self, value: float) -> None:
        num_masters = len(self._widths)
        self._widths.clear()
        self._widths.extend([int(value)] * num_masters)

    @property
    def widths(self) -> list[int]:
        """The "widths for each master. Actually, the width is a representation of the
        angle: `width = round(tan(radians(value)) * 10000)`

        Returns:
            list[int]: The list of "widths"
        """
        # The angle for each master
        return self._widths

    @widths.setter
    def widths(self, value: list[float]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "widths" of class Guide'
        )

    # Methods

    def Transform(self, m: Matrix) -> None:
        """applies Matrix transformation to the Guide

        Args:
            m (Matrix): _description_
        """
        raise NotImplementedError

    def TransformLayer(self, m: Matrix, layernum: int) -> None:
        """applies Matrix transformation to the selected layer of the Guide

        Args:
            m (Matrix): _description_
        """
        raise NotImplementedError



import unittest

from FL.objects.Guide import Guide


class GuideTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        g = Guide()
        assert g.parent is None
        assert g.position == 0
        assert g.width == 21
        assert g.positions == [0] * 16
        assert g.widths == [21] * 16

    def test_angle_45(self) -> None:
        # An angle of 45° is converted to 10000 in the widths list
        g = Guide()
        g.angle = 45
        assert g.width == 10000

    def test_angle_40(self) -> None:
        # An angle of 40° is converted to 8391 in the widths list
        g = Guide()
        g.angle = 40
        assert g.width == 8391
        # ... and converted back with limited precision
        # FIXME: Use actual precision
        # assert g.angle == 40.0000114441
        assert g.angle == 40.000012400755736

    def test_angle_60(self) -> None:
        # An angle of 60° is clamped to 45° (10000 in the widths list)
        g = Guide()
        g.angle = 60
        assert g.width == 10000
        assert g.angle == 45.0

    def test_angle_mm(self) -> None:
        # Setting the angle sets it for all masters
        g = Guide()
        g.angle = 45
        assert g.widths == [10000] * 16

    def test_position_mm(self) -> None:
        # Setting the position sets it for all masters
        g = Guide()
        g.position = 100
        assert g.positions == [100] * 16

    def test_width_mm(self) -> None:
        # Setting the "width" sets it for all masters
        g = Guide()
        g.width = 100
        assert g.widths == [100] * 16



from __future__ import annotations

from FL.fake.Base import Copyable
from FL.objects.Point import Point


class TTHPoint(Copyable):
    """
    TTHPoint - class to represent visual TrueType points

    All coordinates in TTHPoint class are set in 10.6 fixed-point format
    TODO: ^ I think this only applies to TTPoint, not TTHPoint
    """

    __slots__ = ["_x", "_y", "_mode", "_state"]

    # Constructor

    def __init__(
        self,
        tthpoint_or_p_or_x: TTHPoint | Point | int | None = None,
        mode_or_y: int | None = None,
    ) -> None:
        self.x = 0
        self.y = 0
        self._mode = 1
        self._state = 0

        if isinstance(tthpoint_or_p_or_x, TTHPoint):
            self._copy_constructor(tthpoint_or_p_or_x)
        elif isinstance(tthpoint_or_p_or_x, Point):
            # Set from Point
            self.x = tthpoint_or_p_or_x.x
            self.y = tthpoint_or_p_or_x.y
            if mode_or_y is not None:
                self._mode = mode_or_y
        elif isinstance(tthpoint_or_p_or_x, int) or isinstance(
            tthpoint_or_p_or_x, float
        ):
            self.x = float(tthpoint_or_p_or_x)
            if isinstance(mode_or_y, int) or isinstance(mode_or_y, float):
                self.y = float(mode_or_y)
            else:
                raise TypeError

    def __repr__(self) -> str:
        return "<TTHPoint: x: %g, y: %g, mode: %i>" % (self.x, self.y, self.mode)

    # Attributes

    @property
    def x(self) -> float:
        """
        x coordinate

        Returns:
            float: _description_
        """
        return float(self._x)

    @x.setter
    def x(self, value: float) -> None:
        self._x = int(value)

    @property
    def y(self) -> float:
        """
        y coordinate

        Returns:
            float: _description_
        """
        return float(self._y)

    @y.setter
    def y(self, value: float) -> None:
        self._y = int(value)

    @property
    def mode(self) -> int:
        """
        Point's mode (on-curve or off-curve). Read-only.

        Returns:
            int: _description_
        """
        return self._mode

    @mode.setter
    def mode(self, value: int) -> None:
        raise RuntimeError("class TTHPoint has no attribute mode")

    @property
    def state(self) -> int:
        """
        Point's state (touched or not). Read-only.

        Returns:
            int: _description_
        """
        return self._state

    @state.setter
    def state(self, value: int) -> None:
        raise RuntimeError("class TTHPoint has no attribute state")

    # Operations: none

    # Methods: none


"""
Incorrect # of args to:
 TTHPoint(void)
 TTHPoint([TTH]Point[,int newmode])
 THHPoint(int x,int y[,int newmode])

 Point, TTHPoint or Float is expected in arg 1:
 TTHPoint(Point,int mode)
 TTHPoint(TTHPoint,int mode)
 TTHPoint(int X,int Y)

 Point or TTHPoint is expected in arg 1:
 TTHPoint(Point)
 TTHPoint(TTHPoint)

 TTHPoint

 <TTHPoint: x: %g, y: %g, mode: %d> state mode

 class TTHPoint has no attribute %s
"""



from __future__ import annotations

from FL import fl_cmd
from FL.constants import (
    ftFONTLAB,
    ftMACTRUETYPE,
    ftMACTRUETYPE_DFONT,
    ftMACTYPE1,
    ftOPENTYPE,
    ftTRUETYPE,
    ftTYPE1,
    ftTYPE1_MM,
    ftTYPE1ASCII,
    ftTYPE1ASCII_MM,
)
from FL.FLdict import (
    BUTTONCONTROL,
    CHECKBOXCONTROL,
    CHOICECONTROL,
    EDITCONTROL,
    LISTCONTROL,
    OK,
    PREVIEWCONTROL,
    SS_BITMAP,
    SS_BLACKFRAME,
    SS_BLACKRECT,
    SS_CENTER,
    SS_CENTERIMAGE,
    SS_ELLIPSISMASK,
    SS_ENDELLIPSIS,
    SS_ENHMETAFILE,
    SS_ETCHEDFRAME,
    SS_ETCHEDHORZ,
    SS_ETCHEDVERT,
    SS_GRAYFRAME,
    SS_GRAYRECT,
    SS_ICON,
    SS_LEFT,
    SS_LEFTNOWORDWRAP,
    SS_NOPREFIX,
    SS_NOTIFY,
    SS_OWNERDRAW,
    SS_PATHELLIPSIS,
    SS_REALSIZEIMAGE,
    SS_RIGHT,
    SS_RIGHTJUST,
    SS_SIMPLE,
    SS_SUNKEN,
    SS_TYPEMASK,
    SS_USERITEM,
    SS_WHITEFRAME,
    SS_WHITERECT,
    SS_WORDELLIPSIS,
    STATICCONTROL,
    STYLE_BUTTON,
    STYLE_CHECKBOX,
    STYLE_CHOICE,
    STYLE_CUSTOM,
    STYLE_EDIT,
    STYLE_FRAME,
    STYLE_LABEL,
    STYLE_LIST,
    TR_CODE_3D,
    TR_CODE_3DROTATE,
    TR_CODE_ADJUST,
    TR_CODE_ALIGNMENT,
    TR_CODE_ALIGNWIDTH,
    TR_CODE_AUTOHINT,
    TR_CODE_AUTOREPLACE,
    TR_CODE_AUTOSPACING,
    TR_CODE_BEARING,
    TR_CODE_BOLD,
    TR_CODE_COLLEGE,
    TR_CODE_CONVERT2,
    TR_CODE_CONVERT3,
    TR_CODE_DECOMPOSE,
    TR_CODE_DROP_TTH,
    TR_CODE_ENVELOPE,
    TR_CODE_EXTREMES,
    TR_CODE_GRADIENT,
    TR_CODE_MAKELAYER,
    TR_CODE_MIRROR,
    TR_CODE_OPTIMIZE,
    TR_CODE_OVERLAP,
    TR_CODE_RANDOM,
    TR_CODE_REMOVE,
    TR_CODE_REVERSE_ALL,
    TR_CODE_ROTATE,
    TR_CODE_SCALE,
    TR_CODE_SHADOW,
    TR_CODE_SHIFT,
    TR_CODE_SLANT,
    TR_CODE_TOHINTS,
    TR_CODE_TOLINKS,
    TR_CODE_TT_AUTOHINT,
    TR_CODE_WIDTH,
    WS_BORDER,
    WS_CAPTION,
    WS_CHILD,
    WS_CLIPCHILDREN,
    WS_CLIPSIBLINGS,
    WS_DISABLED,
    WS_DLGFRAME,
    WS_GROUP,
    WS_HSCROLL,
    WS_MAXIMIZE,
    WS_MAXIMIZEBOX,
    WS_MINIMIZE,
    WS_MINIMIZEBOX,
    WS_OVERLAPPED,
    WS_OVERLAPPEDWINDOW,
    WS_POPUP,
    WS_POPUPWINDOW,
    WS_SYSMENU,
    WS_TABSTOP,
    WS_THICKFRAME,
    WS_VISIBLE,
    WS_VSCROLL,
    Cancel,
    aALIGN,
    aAUTO,
    aIDENT,
    aIDENT2,
    aIDENT3,
    aNEXT,
    aSAME,
    cBRUSH_NULL,
    cBRUSH_SOLID,
    cDS_AND,
    cDS_BLACK,
    cDS_COPY,
    cDS_NOT,
    cDS_NOTAND,
    cDS_NOTCOPY,
    cDS_NOTOR,
    cDS_NOTXOR,
    cDS_OR,
    cDS_WHITE,
    cDS_XOR,
    cPEN_DASH,
    cPEN_DOT,
    cPEN_NULL,
    cPEN_SOLID,
    cRECT_3DDOWN,
    cRECT_3DDOWN_FLAT,
    cRECT_3DDOWN_SOLID,
    cRECT_3DUP,
    cRECT_3DUP_FLAT,
    cRECT_3DUP_SOLID,
    cRECT_BLACK_FRAME,
    cRECT_BLACK_SOLID,
    cRECT_BLACK_THICK_FRAME,
    cRECT_ELLIPSE,
    cRECT_FOCUS,
    cRECT_GRAY_FRAME,
    cRECT_GRAY_SOLID,
    cRECT_INVERT,
    cRECT_LTGRAY_SOLID,
    cRECT_RECTANGLE,
    cRECT_SIZE,
    cRECT_WHITE_SOLID,
    cRECT_YELLOW_FRAME,
    cRECT_YELLOW_SOLID,
    cRGB_BLACK,
    cRGB_BLUE,
    cRGB_GRAY,
    cRGB_GREEN,
    cRGB_LTGRAY,
    cRGB_RED,
    cRGB_WHITE,
    cRGB_YELLOW,
    cTO_BOTTOM,
    cTO_CENTER,
    cTO_LEFT,
    cTO_RIGHT,
    cTO_TOP,
    cTO_VCENTER,
)
from FL.objects.Anchor import Anchor
from FL.objects.AuditRecord import AuditRecord
from FL.objects.Canvas import Canvas
from FL.objects.Component import Component
from FL.objects.Dialog import Dialog
from FL.objects.Encoding import Encoding
from FL.objects.EncodingRecord import EncodingRecord
from FL.objects.Feature import Feature
from FL.objects.Font import Font
from FL.objects.FontLab import FakeLab
from FL.objects.Glyph import Glyph
from FL.objects.Guide import Guide
from FL.objects.Hint import Hint
from FL.objects.Image import Image
from FL.objects.KerningPair import KerningPair
from FL.objects.Link import Link
from FL.objects.Matrix import Matrix
from FL.objects.NameRecord import NameRecord
from FL.objects.Node import Node, nCURVE, nFIXED, nLINE, nMOVE, nOFF, nSHARP, nSMOOTH
from FL.objects.Options import Options
from FL.objects.Point import Point
from FL.objects.Rect import Rect
from FL.objects.Replace import Replace
from FL.objects.TrueTypeTable import TrueTypeTable
from FL.objects.TTGasp import TTGasp
from FL.objects.TTH import TTH
from FL.objects.TTHCommand import TTHCommand
from FL.objects.TTHPoint import TTHPoint
from FL.objects.TTHProblem import TTHProblem
from FL.objects.TTInfo import TTInfo
from FL.objects.TTPoint import TTPoint
from FL.objects.TTStem import TTStem
from FL.objects.TTVdmx import TTVdmx
from FL.objects.Uni import Uni
from FL.objects.WeightVector import WeightVector

"""
FakeLab. A FontLab Studio 5 replacement for testing Python code.

Everything is only implemented so far as to make FontLab objects importable
outside of FontLab Studio 5, and run tests.
"""

# fl is pre-instantiated
fl = FakeLab()  # noqa: F405

# Everything that is needed for a convincing emulation:

environment = {
    "fl": fl,
    "fl_cmd": fl_cmd,
    # 'fl_decoder': <built-in function raw_unicode_escape_decode>
    # 'fl_encoder': <built-in function raw_unicode_escape_encode>
    # 'fl_stream_reader': <class encodings.raw_unicode_escape.StreamReader>
    # 'fl_stream_writer': <class encodings.raw_unicode_escape.StreamWriter>
    # 'flsys': <module 'flsys' from 'C:\Program Files (x86)\FontLab\Studio 5\flsys.py'>
    # "codecs": codecs,
    # 'cPickle': <module 'cPickle' (built-in)>,
    # "sys": "<module 'sys' (built-in)>",
    # Objects:
    "Anchor": Anchor,
    "AuditRecord": AuditRecord,
    "Canvas": Canvas,
    "Component": Component,
    "Dialog": Dialog,
    "Encoding": Encoding,
    "EncodingRecord": EncodingRecord,
    "Feature": Feature,
    "Font": Font,
    "FontLab": FakeLab,
    "Glyph": Glyph,
    "Guide": Guide,
    "Hint": Hint,
    "Image": Image,
    "KerningPair": KerningPair,
    "Link": Link,
    "Matrix": Matrix,
    "NameRecord": NameRecord,
    "Node": Node,
    "Options": Options,
    "Point": Point,
    "Rect": Rect,
    "Replace": Replace,
    "TrueTypeTable": TrueTypeTable,
    "TTH": TTH,
    "TTHCommand": TTHCommand,
    "TTHPoint": TTHPoint,
    "TTInfo": TTInfo,
    "TTPoint": TTPoint,
    "TTStem": TTStem,
    # Undocumented:
    # 'CmapTable': <built-in function CmapTable>,
    # 'GlyphAudit': <built-in function GlyphAudit>,
    # 'KerningArray': <built-in function KerningArray>,
    # "options": "<FontLab Options>",
    # "RGB": "<function RGB at 0x07109DB0>",
    "TTGasp": TTGasp,
    "TTHProblem": TTHProblem,
    "TTVdmx": TTVdmx,
    "Uni": Uni,
    # "VArray": "<built-in function VArray>",
    # "VExAnchorArray": "<built-in function VExAnchorArray>",
    # "VExCmapTables": "<built-in function VExCmapTables>",
    # "VExFontNames": "<built-in function VExFontNames>",
    # "VExGlyphArray": "<built-in function VExGlyphArray>",
    # "VExGuideArray": "<built-in function VExGuideArray>",
    # "VExHintArray": "<built-in function VExHintArray>",
    # "VExIntArray": "<built-in function VExIntArray>",
    # "VExLinkArray": "<built-in function VExLinkArray>",
    # "VExOTFeatures": "<built-in function VExOTFeatures>",
    # "VExReplaceArray": "<built-in function VExReplaceArray>",
    # "VExSeacArray": "<built-in function VExSeacArray>",
    # "VExTTHCommandArray": "<built-in function VExTTHCommandArray>",
    # "VExTTHPointArray": "<built-in function VExTTHPointArray>",
    # "VExTTStemArray": "<built-in function VExTTStemArray>",
    # "VExTTTables": "<built-in function VExTTTables>",
    # "VSBByteArray": "<built-in function VSBByteArray>,",
    # "VSICharArray": "<built-in function VSICharArray>",
    # "VSITTGaspArray": "<built-in function VSITTGaspArray>",
    # "VSITTVdmxArray": "<built-in function VSITTVdmxArray>",
    # "VSSByteArray": "<built-in function VSSByteArray>",
    # "VSSTTPointArray": "<built-in function VSSTTPointArray>",
    # "VSSWordArray": "<built-in function VSSWordArray>",
    "WeightVector": WeightVector,
    # From this file:
    "ftFONTLAB": ftFONTLAB,
    "ftTYPE1": ftTYPE1,
    "ftTYPE1_MM": ftTYPE1_MM,
    "ftTYPE1ASCII": ftTYPE1ASCII,
    "ftTYPE1ASCII_MM": ftTYPE1ASCII_MM,
    "ftTRUETYPE": ftTRUETYPE,
    "ftOPENTYPE": ftOPENTYPE,
    "ftMACTYPE1": ftMACTYPE1,
    "ftMACTRUETYPE": ftMACTRUETYPE,
    "ftMACTRUETYPE_DFONT": ftMACTRUETYPE_DFONT,
    # From FLdict:
    "cBRUSH_NULL": cBRUSH_NULL,
    "cBRUSH_SOLID": cBRUSH_SOLID,
    "cDS_BLACK": cDS_BLACK,
    "cDS_NOTOR": cDS_NOTOR,
    "cDS_NOTCOPY": cDS_NOTCOPY,
    "cDS_NOT": cDS_NOT,
    "cDS_XOR": cDS_XOR,
    "cDS_NOTAND": cDS_NOTAND,
    "cDS_AND": cDS_AND,
    "cDS_NOTXOR": cDS_NOTXOR,
    "cDS_COPY": cDS_COPY,
    "cDS_OR": cDS_OR,
    "cDS_WHITE": cDS_WHITE,
    "cPEN_DASH": cPEN_DASH,
    "cPEN_DOT": cPEN_DOT,
    "cPEN_NULL": cPEN_NULL,
    "cPEN_SOLID": cPEN_SOLID,
    "cRGB_BLACK": cRGB_BLACK,
    "cRGB_BLUE": cRGB_BLUE,
    "cRGB_GRAY": cRGB_GRAY,
    "cRGB_GREEN": cRGB_GREEN,
    "cRGB_LTGRAY": cRGB_LTGRAY,
    "cRGB_RED": cRGB_RED,
    "cRGB_WHITE": cRGB_WHITE,
    "cRGB_YELLOW": cRGB_YELLOW,
    "EDITCONTROL": EDITCONTROL,
    "STATICCONTROL": STATICCONTROL,
    "CHECKBOXCONTROL": CHECKBOXCONTROL,
    "BUTTONCONTROL": BUTTONCONTROL,
    "LISTCONTROL": LISTCONTROL,
    "CHOICECONTROL": CHOICECONTROL,
    "PREVIEWCONTROL": PREVIEWCONTROL,
    "STYLE_BUTTON": STYLE_BUTTON,
    "STYLE_CHECKBOX": STYLE_CHECKBOX,
    "STYLE_CHOICE": STYLE_CHOICE,
    "STYLE_CUSTOM": STYLE_CUSTOM,
    "STYLE_EDIT": STYLE_EDIT,
    "STYLE_FRAME": STYLE_FRAME,
    "STYLE_LABEL": STYLE_LABEL,
    "STYLE_LIST": STYLE_LIST,
    "cTO_BOTTOM": cTO_BOTTOM,
    "cTO_CENTER": cTO_CENTER,
    "cTO_LEFT": cTO_LEFT,
    "cTO_RIGHT": cTO_RIGHT,
    "cTO_TOP": cTO_TOP,
    "cTO_VCENTER": cTO_VCENTER,
    "aALIGN": aALIGN,
    "aAUTO": aAUTO,
    "aIDENT": aIDENT,
    "aIDENT2": aIDENT2,
    "aIDENT3": aIDENT3,
    "aNEXT": aNEXT,
    "aSAME": aSAME,
    "OK": OK,
    "Cancel": Cancel,
    "nLINE": nLINE,
    "nMOVE": nMOVE,
    "nCURVE": nCURVE,
    "nOFF": nOFF,
    "nSHARP": nSHARP,
    "nSMOOTH": nSMOOTH,
    "nFIXED": nFIXED,
    "TR_CODE_SHIFT": TR_CODE_SHIFT,
    "TR_CODE_MIRROR": TR_CODE_MIRROR,
    "TR_CODE_SCALE": TR_CODE_SCALE,
    "TR_CODE_ROTATE": TR_CODE_ROTATE,
    "TR_CODE_SLANT": TR_CODE_SLANT,
    "TR_CODE_REMOVE": TR_CODE_REMOVE,
    "TR_CODE_AUTOHINT": TR_CODE_AUTOHINT,
    "TR_CODE_TT_AUTOHINT": TR_CODE_TT_AUTOHINT,
    "TR_CODE_AUTOREPLACE": TR_CODE_AUTOREPLACE,
    "TR_CODE_TOLINKS": TR_CODE_TOLINKS,
    "TR_CODE_TOHINTS": TR_CODE_TOHINTS,
    "TR_CODE_DECOMPOSE": TR_CODE_DECOMPOSE,
    "TR_CODE_CONVERT3": TR_CODE_CONVERT3,
    "TR_CODE_CONVERT2": TR_CODE_CONVERT2,
    "TR_CODE_REVERSE_ALL": TR_CODE_REVERSE_ALL,
    "TR_CODE_ALIGNMENT": TR_CODE_ALIGNMENT,
    "TR_CODE_EXTREMES": TR_CODE_EXTREMES,
    "TR_CODE_OVERLAP": TR_CODE_OVERLAP,
    "TR_CODE_WIDTH": TR_CODE_WIDTH,
    "TR_CODE_BEARING": TR_CODE_BEARING,
    "TR_CODE_ALIGNWIDTH": TR_CODE_ALIGNWIDTH,
    "TR_CODE_AUTOSPACING": TR_CODE_AUTOSPACING,
    "TR_CODE_BOLD": TR_CODE_BOLD,
    "TR_CODE_COLLEGE": TR_CODE_COLLEGE,
    "TR_CODE_SHADOW": TR_CODE_SHADOW,
    "TR_CODE_3D": TR_CODE_3D,
    "TR_CODE_3DROTATE": TR_CODE_3DROTATE,
    "TR_CODE_GRADIENT": TR_CODE_GRADIENT,
    "TR_CODE_RANDOM": TR_CODE_RANDOM,
    "TR_CODE_ENVELOPE": TR_CODE_ENVELOPE,
    "TR_CODE_MAKELAYER": TR_CODE_MAKELAYER,
    "TR_CODE_OPTIMIZE": TR_CODE_OPTIMIZE,
    "TR_CODE_DROP_TTH": TR_CODE_DROP_TTH,
    "TR_CODE_ADJUST": TR_CODE_ADJUST,
    "SS_BITMAP": SS_BITMAP,
    "SS_BLACKFRAME": SS_BLACKFRAME,
    "SS_BLACKRECT": SS_BLACKRECT,
    "SS_CENTER": SS_CENTER,
    "SS_CENTERIMAGE": SS_CENTERIMAGE,
    "SS_ELLIPSISMASK": SS_ELLIPSISMASK,
    "SS_ENDELLIPSIS": SS_ENDELLIPSIS,
    "SS_ENHMETAFILE": SS_ENHMETAFILE,
    "SS_ETCHEDFRAME": SS_ETCHEDFRAME,
    "SS_ETCHEDHORZ": SS_ETCHEDHORZ,
    "SS_ETCHEDVERT": SS_ETCHEDVERT,
    "SS_GRAYFRAME": SS_GRAYFRAME,
    "SS_GRAYRECT": SS_GRAYRECT,
    "SS_ICON": SS_ICON,
    "SS_LEFT": SS_LEFT,
    "SS_LEFTNOWORDWRAP": SS_LEFTNOWORDWRAP,
    "SS_NOPREFIX": SS_NOPREFIX,
    "SS_NOTIFY": SS_NOTIFY,
    "SS_OWNERDRAW": SS_OWNERDRAW,
    "SS_PATHELLIPSIS": SS_PATHELLIPSIS,
    "SS_REALSIZEIMAGE": SS_REALSIZEIMAGE,
    "SS_RIGHT": SS_RIGHT,
    "SS_RIGHTJUST": SS_RIGHTJUST,
    "SS_SIMPLE": SS_SIMPLE,
    "SS_SUNKEN": SS_SUNKEN,
    "SS_TYPEMASK": SS_TYPEMASK,
    "SS_USERITEM": SS_USERITEM,
    "SS_WHITEFRAME": SS_WHITEFRAME,
    "SS_WHITERECT": SS_WHITERECT,
    "SS_WORDELLIPSIS": SS_WORDELLIPSIS,
    "WS_BORDER": WS_BORDER,
    "WS_CAPTION": WS_CAPTION,
    "WS_CHILD": WS_CHILD,
    "WS_CLIPCHILDREN": WS_CLIPCHILDREN,
    "WS_CLIPSIBLINGS": WS_CLIPSIBLINGS,
    "WS_DISABLED": WS_DISABLED,
    "WS_DLGFRAME": WS_DLGFRAME,
    "WS_GROUP": WS_GROUP,
    "WS_HSCROLL": WS_HSCROLL,
    "WS_MAXIMIZE": WS_MAXIMIZE,
    "WS_MAXIMIZEBOX": WS_MAXIMIZEBOX,
    "WS_MINIMIZE": WS_MINIMIZE,
    "WS_MINIMIZEBOX": WS_MINIMIZEBOX,
    "WS_OVERLAPPED": WS_OVERLAPPED,
    "WS_OVERLAPPEDWINDOW": WS_OVERLAPPEDWINDOW,
    "WS_POPUP": WS_POPUP,
    "WS_POPUPWINDOW": WS_POPUPWINDOW,
    "WS_SYSMENU": WS_SYSMENU,
    "WS_TABSTOP": WS_TABSTOP,
    "WS_THICKFRAME": WS_THICKFRAME,
    "WS_VISIBLE": WS_VISIBLE,
    "WS_VSCROLL": WS_VSCROLL,
    "cRECT_3DDOWN": cRECT_3DDOWN,
    "cRECT_3DDOWN_FLAT": cRECT_3DDOWN_FLAT,
    "cRECT_3DDOWN_SOLID": cRECT_3DDOWN_SOLID,
    "cRECT_3DUP": cRECT_3DUP,
    "cRECT_3DUP_FLAT": cRECT_3DUP_FLAT,
    "cRECT_3DUP_SOLID": cRECT_3DUP_SOLID,
    "cRECT_BLACK_FRAME": cRECT_BLACK_FRAME,
    "cRECT_BLACK_SOLID": cRECT_BLACK_SOLID,
    "cRECT_BLACK_THICK_FRAME": cRECT_BLACK_THICK_FRAME,
    "cRECT_ELLIPSE": cRECT_ELLIPSE,
    "cRECT_FOCUS": cRECT_FOCUS,
    "cRECT_GRAY_FRAME": cRECT_GRAY_FRAME,
    "cRECT_GRAY_SOLID": cRECT_GRAY_SOLID,
    "cRECT_INVERT": cRECT_INVERT,
    "cRECT_LTGRAY_SOLID": cRECT_LTGRAY_SOLID,
    "cRECT_RECTANGLE": cRECT_RECTANGLE,
    "cRECT_SIZE": cRECT_SIZE,
    "cRECT_WHITE_SOLID": cRECT_WHITE_SOLID,
    "cRECT_YELLOW_FRAME": cRECT_YELLOW_FRAME,
    "cRECT_YELLOW_SOLID": cRECT_YELLOW_SOLID,
    # Undocumented:
    "CMAPTYPE_CODEPAGE": 1,
    "CMAPTYPE_FONTWINDOW": 2,
    "CMAPTYPE_UNICODE": 0,
    "metricsKERNING": 2,
    "metricsMETRICS": 1,
    "metricsSAMPLE": 0,
    "metricsTEXT": 3,
    "mfMARK_RESULT": 2,
    "mfOVERLAP_IGNORE": 1,
    "mfOVERLAP_RENAME_DEST": 16,
    "mfOVERLAP_RENAME_SOURCE": 8,
    "mfOVERLAP_REPLACE": 32,
    "mfOVERLAP_SKIP": 4,
    "STYLE_BEVELBUTTON": 268435456,
    "TTFLAGS_CREATE_CHARACTERS": 262144,
    "TTFLAGS_CREATE_HDMX": 524288,
    "TTFLAGS_CREATE_VDMX": 131072,
    "TTFLAGS_CREATE_VMTX": 1048576,
    # TODO: tools, e.g.:
    # 'Circles': <module 'Circles' from 'C:\Users\jens\Documents\FontLab\Studio 5\macros\system\tool\Circles.pyc'>
    # 'Drop': <module 'Drop' from 'C:\Users\jens\Documents\FontLab\Studio 5\macros\system\tool\Drop.pyc'>
    # 'Line': <module 'Line' from 'C:\Users\jens\Documents\FontLab\Studio 5\macros\system\tool\Line.pyc'>
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph
    from FL.objects.Image import Image
    from FL.objects.Point import Point
    from FL.objects.Rect import Rect
    from FL.objects.WeightVector import WeightVector


class Canvas(Copyable):
    """
    Canvas - class to represent screen paint area and operations

    This class can perform automatic coordinates transformations, so you can draw in
    Glyph coordinate space or, if you prefer, in screen coordinates directly.

    Use constants defined for FontLab module in `Fldict` to assign attributes to
    `Canvas` object
    """

    __slots__ = [
        "_bk_color",
        "_bk_mode",
        "_brush_color",
        "_brush_style",
        "_delta",
        "_draw_style",
        "_pen_color",
        "_pen_style",
        "_scale",
        "_text_color",
    ]

    # Constructor

    def __init__(self, canvas: Canvas | None = None) -> None:
        """
        Never create Canvas object explicitly - they must be obtained from the FontLab's
        window classes

        Canvas()       - generic constructor, creates an empty Canvas
        Canvas(Canvas) - copy constructor
        """
        self._bk_color = 0
        self._bk_mode = 0
        self._brush_color = 0
        self._brush_style = 0
        self._delta = Point()
        self._draw_style = 0
        self._pen_color = 0
        self._pen_style = 0
        self._scale = Point()
        self._text_color = 0

        if isinstance(canvas, Canvas):
            self._copy_constructor(canvas)

    # Attributes

    @property
    def delta(self) -> Point:
        """
        Horizontal shift during coordinates transformation

        Returns:
            Point: _description_
        """
        return self._delta

    @property
    def scale(self) -> Point:
        """
        Scaling during coordinates transformation

        Returns:
            Point: _description_
        """
        return self._scale

    @property
    def parent(self) -> Any:
        """
        parent object

        Returns:
            Any: The parent
        """
        raise NotImplementedError

    @property
    def draw_style(self) -> None:
        raise AttributeError

    @draw_style.setter
    def draw_style(self, value: int) -> None:
        """
        Sets current drawing mode (copy, XOR, invert etc.)

        Args:
            value (int): _description_
        """
        self._draw_style = value

    @property
    def pen_color(self) -> None:
        raise AttributeError

    @pen_color.setter
    def pen_color(self, value: int) -> None:
        """
        Sets current pen color

        Args:
            value (int): _description_
        """
        self._pen_color = value

    @property
    def pen_style(self) -> None:
        raise AttributeError

    @pen_style.setter
    def pen_style(self, value: int) -> None:
        """
        Sets current pen style (solid, dashed etc.)

        Args:
            value (int): _description_
        """
        self._pen_style = value

    @property
    def brush_color(self) -> None:
        raise AttributeError

    @brush_color.setter
    def brush_color(self, value: int) -> None:
        """
        Sets current brush color

        Args:
            value (int): _description_
        """
        self._brush_color = value

    @property
    def brush_style(self) -> None:
        raise AttributeError

    @brush_style.setter
    def brush_style(self, value: int) -> None:
        """
        Sets current brush style (solid, patterned etc.)

        Args:
            value (int): _description_
        """
        self._brush_style = value

    @property
    def text_color(self) -> None:
        raise AttributeError

    @text_color.setter
    def text_color(self, value: int) -> None:
        """
        Sets color for a text

        Args:
            value (int): _description_
        """
        self._text_color = value

    @property
    def bk_color(self) -> None:
        raise AttributeError

    @bk_color.setter
    def bk_color(self, value: int) -> None:
        """
        Sets color for a text background

        Args:
            value (int): _description_
        """
        self._bk_color = value

    @property
    def bk_mode(self) -> None:
        raise AttributeError

    @bk_mode.setter
    def bk_mode(self, value: int) -> None:
        """
        Sets background mode for a text output (transparent or opaque)

        Args:
            value (int): _description_
        """
        self._bk_mode = value

    # Methods

    def MoveTo(self, p_or_x: Point | float, y: float | None = None) -> None:
        """
        Moves current position to p or (x, y) coordinates

        Args:
            p_or_x (Point | float): _description_
            y (float | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def LineTo(self, p_or_x: Point | float, y: float | None = None) -> None:
        """
        Draws a straight line to the position p or (x, y) coordinates

        Args:
            p_or_x (Point | float): _description_
            y (float | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def CurveTo(self, p0: Point, p1: Point, p2: Point) -> None:
        """
        Draws a 3th-order Bezier curve from the current point to points p0, p1, p2

        Args:
            p0 (Point): _description_
            p1 (Point): _description_
            p2 (Point): _description_
        """
        raise NotImplementedError

    def SplineTo(self, p0: Point, p1: Point) -> None:
        """
        Draws a 2th-order Bezier curve from the current point to points p0, p1

        Args:
            p0 (Point): _description_
            p1 (Point): _description_
        """
        raise NotImplementedError

    def Ellipse(
        self,
        r_or_p0_or_x0: Rect | Point | float,
        p1_or_y0: Point | float | None = None,
        x1: float | None = None,
        y1: float | None = None,
    ) -> None:
        """
        Draws a filled ellipse defined by the Rect r, points p0 and p1 or set of
        coordinates

        Args:
            r_or_p0_or_x0 (Rect | Point | float): _description_
            p1_or_y0 (Point | float | None, optional): _description_. Defaults to None.
            x1 (float | None, optional): _description_. Defaults to None.
            y1 (float | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def Rectangle(
        self,
        r_or_p0_or_x0: Rect | Point | float,
        p1_or_y0: Point | float | None = None,
        x1: float | None = None,
        y1: float | None = None,
    ) -> None:
        """
        Draws a filled rectangle defined by the Rect r, points p0 and p1 or set of
        coordinates

        Args:
            r_or_p0_or_x0 (Rect | Point | float): _description_
            p1_or_y0 (Point | float | None, optional): _description_. Defaults to None.
            x1 (float | None, optional): _description_. Defaults to None.
            y1 (float | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def Convert(self, p: Point):
        """
        Converts coordinates from the source coordinate space

        Args:
            p (Point): _description_
        """
        raise NotImplementedError

    def UnConvert(self, p: Point):
        """
        Converts screen coordinates to the currently defined coordinate space

        Args:
            p (Point): _description_
        """
        raise NotImplementedError

    def FitGlyph(self, r: Rect, g: Glyph, w: WeightVector | None = None) -> None:
        """
        Recalculates parameters of coordinate conversion function to fit glyph g into
        rectangle r using currently selected options.

        Args:
            r (Rect): _description_
            g (Glyph): _description_
            w (WeightVector): _description_
        """
        raise NotImplementedError

    def FillGlyph(self, g: Glyph, w: WeightVector | None = None) -> None:
        """
        Fills the glyph g using current coordinate transformation.

        Args:
            g (Glyph): _description_
            w (WeightVector | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def OutlineGlyph(self, g: Glyph, w: WeightVector | None = None) -> None:
        """
        Draws the glyph g outline using current coordinate transformation.

        Args:
            g (Glyph): _description_
            w (WeightVector | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def TextOut(
        self, p_or_x: Point | float, s_or_y: str | float, s: str | None = None
    ) -> None:
        """
        Draws the string s in the position p or (x, y)

        Args:
            p_or_x (Point | float): _description_
            s_or_y (str | float): _description_
            s (str | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def PutImage(self, i: Image, p: Point, mode: int | None = None) -> None:
        """
        Puts Image i at the selected point

        Args:
            i (Image): _description_
            p (Point): _description_
            mode (int | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable
from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Font import Font
    from FL.objects.Glyph import Glyph


class Component(Copyable):
    """
    Component - class to represent glyph component

    Components are defined by the glyph index which they reference,
    Shift of components origin point and scale factor of a component
    Scale is measured in relation to 1.0, so 100% scale is 1.0 and 60% - 0.6
    This class is Multiple Master - compatible
    """

    __slots__ = ["_deltas", "_scales", "_index", "_parent"]

    # Constructor

    def __init__(
        self,
        component_or_index: Component | int | None = None,
        delta: Point | None = None,
        scale: Point | None = None,
    ) -> None:
        """
        Component()          - generic constructor, creates an empty Component
        Component(Component) - copy constructor
        Component(index)     - creates component referencing
                               glyph index with zero shift and 100% scale
        Component(index, Point(integer) delta)
                             - creates component referencing
                               glyph index with delta shift and 100% scale
        Component(index, Point(integer) delta, Point(float) scale)
                             - creates component referencing
                               glyph index with delta shift
                               and scale factor defined by scale

        Args:
            component_or_index (Component | int | None, optional): _description_. Defaults to None.
            delta (Point | None, optional): _description_. Defaults to None.
            scale (Point | None, optional): _description_. Defaults to None.
        """
        # Init with max num masters and -1 reference glyph
        self._deltas = [Point(0, 0)] * 16
        self._scales = [Point(1.0, 1.0)] * 16
        self._index = -1
        self._parent: Glyph | None = None

        if isinstance(component_or_index, Component):
            self._copy_constructor(component_or_index)

        elif isinstance(component_or_index, int):
            self._index = component_or_index
            if delta is not None:
                self.delta = delta
                if scale is not None:
                    self.scale = scale
        # else: Empty Component

    # Additions for FakeLab

    def fake_deserialize(
        self, num_masters: int, data: dict[str, int | list[float]]
    ) -> None:
        self.index = data.get("gid", -1)
        self._deltas = [
            Point(data["offsetX"][i], data["offsetY"][i]) for i in range(num_masters)
        ]
        self._scales = [
            Point(data["scaleX"][i], data["scaleY"][i]) for i in range(num_masters)
        ]

    def fake_serialize(self) -> dict[str, int | list[int | float]]:
        d = {
            "gid": self.index,
            "offsetX": [int(p.x) for p in self.deltas],  # FIXME: Round or int?
            "offsetY": [int(p.y) for p in self.deltas],  # FIXME: Round or int?
            "scaleX": [p.x for p in self.scales],
            "scaleY": [p.y for p in self.scales],
        }
        return d

    # Attributes

    @property
    def parent(self) -> Glyph | None:
        """
        parent object, Glyph
        """
        return self._parent

    @property
    def index(self) -> int:
        """
        referencing glyph index
        """
        return self._index

    @index.setter
    def index(self, value: int) -> None:
        self._index = value

    @property
    def delta(self) -> Point:
        # TODO: What is the point's parent?
        return self._deltas[0]

    @delta.setter
    def delta(self, value: Point) -> None:
        """
        Shift value

        When setting the shift through this method, the value is used for all masters.

        Args:
            value (Point): _description_
        """
        self._deltas = [Point(int(value.x), int(value.y))] * len(self._deltas)

    @property
    def scale(self) -> Point:
        """
        scale factor
        """
        # TODO: What is the point's parent?
        return self._scales[0]

    @scale.setter
    def scale(self, value: Point) -> None:
        """
        scale factor

        Args:
            value (Point): _description_

        When setting the scale through this method, the value is used for all masters.
        """
        # TODO: What is the point's parent?
        self._scales = [Point(value.x, value.y)] * len(self._scales)

    @property
    def deltas(self) -> list[Point]:
        """
        list of shift values for each master

        Setting the list, or only one item of the list doesn't work in FL.
        You need to get one of the points and modify it directly
        """
        # TODO: parent of the points must be the deltas list
        return self._deltas

    @deltas.setter
    def deltas(self, value: list[Point]) -> None:
        # Setting the list, or only one item of the list doesn't work in FL.
        # You need to get one of the points and modify it directly
        raise RuntimeError(
            'Attempt to write read only attribute "deltas" of class Component'
        )

    @property
    def scales(self) -> list[Point]:
        """
        list of scale values for each master

        Setting the list, or only one item of the list doesn't work in FL.
        You need to get one of the points and modify it directly
        """
        # TODO: parent of the points must be the scales list
        return self._scales

    @scales.setter
    def scales(self, value: list[Point]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "scales" of class Component'
        )

    # Operations: none

    # Methods

    def Get(self, f: Font | None = None) -> Glyph:
        """
        Creates a glyph from component applying delta and scale transformations.

        Font parameter is not needed when component has a parent
        FIXME: Apparently, the font parameter is always needed

        Args:
            f (Font | None, optional): _description_. Defaults to None.
        """
        if f is None:
            raise RuntimeError(
                "In order to get glyph using component created from orphan glyph "
                "specify source font"
            )
        raise NotImplementedError

    def Paste(self) -> None:
        """
        Appends component to a parent glyph as a set of outlines.

        Component must have a parent
        """
        raise NotImplementedError



from __future__ import annotations

from collections import UserList
from typing import TYPE_CHECKING

from FL.fake.Base import Copyable
from FL.objects.EncodingRecord import EncodingRecord

if TYPE_CHECKING:
    from FL.objects.Font import Font


class Encoding(UserList[EncodingRecord], Copyable):
    """
    Encoding - class to represent Encoding
    """

    __slots__ = ["_parent", "data"]

    # Constructor

    def __init__(self, encoding_or_none: Encoding | None = None) -> None:
        """
        Encoding()         - generic constructor, creates Encoding
        Encoding(Encoding) - copy constructor

        Args:
            encoding_or_none (Encoding | None, optional): _description_. Defaults to None.
        """
        self._parent: Font | None = None
        self.data: list[EncodingRecord] = []
        if isinstance(encoding_or_none, Encoding):
            self._copy_constructor(encoding_or_none)
        # ??? We can't inspect a just initialized encoding in FL, it just crashes.
        # else:
        #     self._set_default()

    def __repr__(self) -> str:
        if self.parent is None:
            p = "orpahn"  # sic
        else:
            p = f'parent: "{self.parent.font_name or "(null)"}"'
        return f"<Encoding: {p}>"

    # Attributes

    @property
    def parent(self) -> Font | None:
        """
        Encoding's parent object, Font
        """
        return self._parent

    # Operations

    def __delitem__(self, i: int) -> None:
        """
        del Encoding[] - remove an element from the encoding
        """
        del self.data[i]

    def __getitem__(self, i: int) -> EncodingRecord:
        """
        Accesses individial EncodingRecord objects
        """
        return self.data[i]

    def __len__(self) -> int:
        """
        Return the number of EncodingRecords in the Encoding.
        """
        return len(self.data)

    # Methods

    def append(self, encoding_record: EncodingRecord) -> None:
        """
        Append an EncodingRecord to the end of the encoding.

        Args:
            encoding_record (EncodingRecord): _description_
        """
        self.data.append(encoding_record)

    def insert(self, index: int, encoding_record: EncodingRecord) -> None:
        """
        Insert an EncodingRecord at index.

        Args:
            index (int): _description_
            encoding_record (EncodingRecord): _description_
        """
        self.data.insert(index, encoding_record)

    def FillUnencoded(self) -> None:
        raise NotImplementedError

    def FillUnicodes(self) -> None:
        raise NotImplementedError

    def FindName(self, name: str) -> int:
        """
        Find a glyph name in the encoding and return its index or -1.

        Args:
            name (str): The glyph name
        """
        raise NotImplementedError

    def Load(self, filename: str) -> None:
        """
        Opens encoding from .ENC format.

        Args:
            filename (str): _description_
        """
        raise NotImplementedError

    def Save(self, filename: str, EncodingTitle: str, Id: int) -> None:
        """
        Saves encoding in .ENC format.

        Args:
            filename (str): _description_
            EncodingTitle (str): _description_
            Id (int): _description_
        """
        raise NotImplementedError

    # Internal

    def load_font_default(self) -> None:
        # How is the unicode derived? It is not stored in Encoding in the VFB.
        for name, uni in (
            ("_0000", -1),
            ("_0001", -1),
            ("_0002", -1),
            ("_0003", -1),
            ("_0004", -1),
            ("breve", 728),
            ("dotaccent", 729),
            ("_0007", -1),
            ("ring", 730),
            ("_0009", -1),
            ("hungarumlaut", 733),
            ("ogonek", 731),
            ("caron", 711),
            ("dotlessi", 305),
            ("_0014", -1),
            ("_0015", -1),
            ("_0016", -1),
            ("_0017", -1),
            ("_0018", -1),
            ("_0019", -1),
            ("_0020", -1),
            ("_0021", -1),
            ("_0022", -1),
            ("_0023", -1),
            ("_0024", -1),
            ("fraction", 8260),
            ("fi", 64257),
            ("fl", 64258),
            ("Lslash", 321),
            ("lslash", 322),
            ("Zcaron", 381),
            ("zcaron", 382),
            ("space", 32),
            ("exclam", 33),
            ("quotedbl", 34),
            ("numbersign", 35),
            ("dollar", 36),
            ("percent", 37),
            ("ampersand", 38),
            ("quotesingle", 39),
            ("parenleft", 40),
            ("parenright", 41),
            ("asterisk", 42),
            ("plus", 43),
            ("comma", 44),
            ("hyphen", 45),
            ("period", 46),
            ("slash", 47),
            ("zero", 48),
            ("one", 49),
            ("two", 50),
            ("three", 51),
            ("four", 52),
            ("five", 53),
            ("six", 54),
            ("seven", 55),
            ("eight", 56),
            ("nine", 57),
            ("colon", 58),
            ("semicolon", 59),
            ("less", 60),
            ("equal", 61),
            ("greater", 62),
            ("question", 63),
            ("at", 64),
            ("A", 65),
            ("B", 66),
            ("C", 67),
            ("D", 68),
            ("E", 69),
            ("F", 70),
            ("G", 71),
            ("H", 72),
            ("I", 73),
            ("J", 74),
            ("K", 75),
            ("L", 76),
            ("M", 77),
            ("N", 78),
            ("O", 79),
            ("P", 80),
            ("Q", 81),
            ("R", 82),
            ("S", 83),
            ("T", 84),
            ("U", 85),
            ("V", 86),
            ("W", 87),
            ("X", 88),
            ("Y", 89),
            ("Z", 90),
            ("bracketleft", 91),
            ("backslash", 92),
            ("bracketright", 93),
            ("asciicircum", 94),
            ("underscore", 95),
            ("grave", 96),
            ("a", 97),
            ("b", 98),
            ("c", 99),
            ("d", 100),
            ("e", 101),
            ("f", 102),
            ("g", 103),
            ("h", 104),
            ("i", 105),
            ("j", 106),
            ("k", 107),
            ("l", 108),
            ("m", 109),
            ("n", 110),
            ("o", 111),
            ("p", 112),
            ("q", 113),
            ("r", 114),
            ("s", 115),
            ("t", 116),
            ("u", 117),
            ("v", 118),
            ("w", 119),
            ("x", 120),
            ("y", 121),
            ("z", 122),
            ("braceleft", 123),
            ("bar", 124),
            ("braceright", 125),
            ("asciitilde", 126),
            ("_0127", -1),
            ("Euro", 8364),
            ("_0129", -1),
            ("quotesinglbase", 8218),
            ("florin", 402),
            ("quotedblbase", 8222),
            ("ellipsis", 8230),
            ("dagger", 8224),
            ("daggerdbl", 8225),
            ("circumflex", 710),
            ("perthousand", 8240),
            ("Scaron", 352),
            ("guilsinglleft", 8249),
            ("OE", 338),
            ("_0141", -1),
            ("_0142", -1),
            ("_0143", -1),
            ("_0144", -1),
            ("quoteleft", 8216),
            ("quoteright", 8217),
            ("quotedblleft", 8220),
            ("quotedblright", 8221),
            ("bullet", 8226),
            ("endash", 8211),
            ("emdash", 8212),
            ("tilde", 732),
            ("trademark", 8482),
            ("scaron", 353),
            ("guilsinglright", 8250),
            ("oe", 339),
            ("_0157", -1),
            ("_0158", -1),
            ("Ydieresis", 376),
            ("uni00A0", 160),
            ("exclamdown", 161),
            ("cent", 162),
            ("sterling", 163),
            ("currency", 164),
            ("yen", 165),
            ("brokenbar", 166),
            ("section", 167),
            ("dieresis", 168),
            ("copyright", 169),
            ("ordfeminine", 170),
            ("guillemotleft", 171),
            ("logicalnot", 172),
            ("minus", 8722),
            ("registered", 174),
            ("macron", 175),
            ("degree", 176),
            ("plusminus", 177),
            ("twosuperior", 178),
            ("threesuperior", 179),
            ("acute", 180),
            ("mu", 181),
            ("paragraph", 182),
            ("periodcentered", 183),
            ("cedilla", 184),
            ("onesuperior", 185),
            ("ordmasculine", 186),
            ("guillemotright", 187),
            ("onequarter", 188),
            ("onehalf", 189),
            ("threequarters", 190),
            ("questiondown", 191),
            ("Agrave", 192),
            ("Aacute", 193),
            ("Acircumflex", 194),
            ("Atilde", 195),
            ("Adieresis", 196),
            ("Aring", 197),
            ("AE", 198),
            ("Ccedilla", 199),
            ("Egrave", 200),
            ("Eacute", 201),
            ("Ecircumflex", 202),
            ("Edieresis", 203),
            ("Igrave", 204),
            ("Iacute", 205),
            ("Icircumflex", 206),
            ("Idieresis", 207),
            ("Eth", 208),
            ("Ntilde", 209),
            ("Ograve", 210),
            ("Oacute", 211),
            ("Ocircumflex", 212),
            ("Otilde", 213),
            ("Odieresis", 214),
            ("multiply", 215),
            ("Oslash", 216),
            ("Ugrave", 217),
            ("Uacute", 218),
            ("Ucircumflex", 219),
            ("Udieresis", 220),
            ("Yacute", 221),
            ("Thorn", 222),
            ("germandbls", 223),
            ("agrave", 224),
            ("aacute", 225),
            ("acircumflex", 226),
            ("atilde", 227),
            ("adieresis", 228),
            ("aring", 229),
            ("ae", 230),
            ("ccedilla", 231),
            ("egrave", 232),
            ("eacute", 233),
            ("ecircumflex", 234),
            ("edieresis", 235),
            ("igrave", 236),
            ("iacute", 237),
            ("icircumflex", 238),
            ("idieresis", 239),
            ("eth", 240),
            ("ntilde", 241),
            ("ograve", 242),
            ("oacute", 243),
            ("ocircumflex", 244),
            ("otilde", 245),
            ("odieresis", 246),
            ("divide", 247),
            ("oslash", 248),
            ("ugrave", 249),
            ("uacute", 250),
            ("ucircumflex", 251),
            ("udieresis", 252),
            ("yacute", 253),
            ("thorn", 254),
            ("ydieresis", 255),
        ):
            rec = EncodingRecord()
            rec.name = name
            rec.unicode = uni
            self.append(rec)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from vfbLib.enum import F, G, M, T
from vfbLib.vfb.vfb import Vfb

from FL.helpers.nametables import StandardNametable
from FL.objects.Encoding import Encoding
from FL.objects.EncodingRecord import EncodingRecord
from FL.objects.Glyph import Glyph
from FL.objects.NameRecord import NameRecord

if TYPE_CHECKING:
    from pathlib import Path

    from FL.objects.Font import Font


logger = logging.getLogger(__name__)


font_mapping_direct = {
    F.font_name,
    F.weight_vector,
    F.unique_id,
    F.version,
    F.notice,
    F.full_name,
    F.family_name,
    F.pref_family_name,
    F.menu_name,
    F.apple_name,
    F.weight,
    F.width,
    F.copyright,
    F.trademark,
    F.designer,
    F.designer_url,
    F.vendor_url,
    F.source,
    F.is_fixed_pitch,
    F.weight_code,
    F.italic_angle,
    F.slant_angle,
    F.underline_position,
    F.underline_thickness,
    F.ms_charset,
    F.panose,
    F.tt_version,
    F.tt_u_id,
    F.style_name,
    F.pref_style_name,
    F.mac_compatible,
    F.vendor,
    # F.xuid,  # has no setter
    F.xuid_num,
    F.year,
    F.version_major,
    F.version_minor,
    F.upm,
    F.fond_id,
    F.blue_values_num,
    F.other_blues_num,
    F.family_blues_num,
    F.family_other_blues_num,
    F.stem_snap_h_num,
    F.stem_snap_v_num,
    F.font_style,
    F.pcl_id,
    F.vp_id,
    F.ms_id,
    F.pcl_chars_set,
    # F.unicoderanges,  # must be converted to a list of bits
    F.note,
    F.customdata,
    F.default_character,
}

glyph_mapping = {
    G.Links,
    G.image,
    G.Bitmaps,
    G.E2023,
    G.Sketch,
    G.HintingOptions,
    G.mask,
    G.MaskMetrics,
    G.MaskMetricsMM,
    G.Origin,
    G.unicodes,
    G.UnicodesNonBMP,
    G.mark,
    G.customdata,
    G.note,
    G.GDEFData,
    G.AnchorsProperties,
    G.AnchorsMM,
    G.GuideProperties,
}

ttinfo_mapping_direct = {
    T.hhea_line_gap,
    T.hhea_ascender,
    T.hhea_descender,
}


class VfbToFontReader:
    """
    Instantiate the Font object and all related objects from a `vfbLib.vfb.vfb.Vfb` (I
    know...) object (low-level representation of the binary VFB format)
    """

    def __init__(self, vfb_path: Path, font: Font) -> None:
        """
        Load the VFB file from `vfb_path` and instantiate its data into `font`.

        Args:
            vfb_path (Path): The file path from which to load the VFB data
            font (Font): The target object of the data
        """
        self.vfb_path = vfb_path
        self.font = font
        self.nametable = StandardNametable()
        self.open_vfb()
        self.read_into_font()

    def open_vfb(self) -> None:
        self.vfb = Vfb(self.vfb_path)
        self.vfb.decompile()

    def read_into_font(self) -> None:
        classes: list[str] = []
        glyph: Glyph | None = None
        gids = {}

        font = self.font
        font.fake_clear_defaults()

        for e in self.vfb.entries:
            key = e.id
            assert isinstance(key, int)
            data = e.data

            if key in font_mapping_direct:
                attr = F(key).name
                if hasattr(font, attr):
                    setattr(font, attr, data)
                else:
                    raise AttributeError(f"Unknown font attribute: {attr}")
                continue

            if key in glyph_mapping:
                assert glyph is not None, "Glyph must exist before adding data"
                glyph.fake_deserialize(key, data)
                continue

            if key in (T.cvt, T.prep, T.fpgm):
                font.ttinfo.fake_set_binary(T(key).name, data)
                continue

            if key in (
                T.TrueTypeStems,
                T.TrueTypeStemPPEMs1,
                T.TrueTypeStemPPEMs2And3,
                T.TrueTypeStemPPEMs,
            ):
                font.ttinfo.fake_deserialize_stems(data)
                continue

            if key in (
                F.E257,
                F.E271,
                F.E513,
                F.E518,
                F.E527,
                F.E1068,
                F.E1140,
                F.E1502,
                F.E2030,
                F.E2030,
            ):
                font._unknown_pleasures[key] = data
                continue

            if key in ttinfo_mapping_direct:
                setattr(font.ttinfo, T(key).name, data)
                continue

            if key in (T.E1604, T.E2032):
                font.ttinfo._unknown_pleasures[key] = data
                continue

            match key:
                case F.xuid:
                    font._xuid = data
                case F.Encoding:
                    gid, glyph_name = data
                    gids[gid] = glyph_name
                case F.EncodingDefault:
                    # Where is this used?
                    gid, glyph_name = data
                    e = EncodingRecord()
                    e.name = glyph_name
                    font._encoding_default.append(e)
                case F.MasterCount:
                    font._masters_count = data
                case F.License:
                    font._license = data
                case F.LicenseURL:
                    font._license_url = data
                case F.PostScriptHintingOptions:
                    font._postscript_hinting_options = data
                case T.gasp:
                    font.ttinfo.fake_deserialize_gasp(data)
                case F.ttinfo:
                    font.ttinfo.fake_deserialize(data)
                case T.vdmx:
                    font.ttinfo.fake_deserialize_vdmx(data)
                case T.TrueTypeZones:
                    font.ttinfo.fake_deserialize_zones(data)
                case F.unicoderanges:
                    font.unicoderanges = data
                case T.stemsnaplimit:
                    font.ttinfo._stemsnaplimit = data
                case T.zoneppm:
                    font.ttinfo._zoneppm = data
                case T.codeppm:
                    font.ttinfo._codeppm = data
                case T.TrueTypeZoneDeltas:
                    font.ttinfo.fake_deserialize_zone_deltas(data)
                case F.fontnames:
                    assert isinstance(data, list)
                    for nr in data:
                        font.fontnames.append(NameRecord(tuple(nr)))
                case F.CustomCMAPs:
                    font._custom_cmaps = data
                case F.PCLTTable:
                    font._pclt_table = data
                case F.ExportPCLTTable:
                    font._export_pclt_table = data
                case F.TrueTypeTable:
                    font.truetypetables.append(data)
                case F.MetricsClassFlags:
                    font._metrics_class_flags = data
                case F.KerningClassFlags:
                    font._kerning_class_flags = data
                case F.features:
                    font.fake_deserialize_features(data)
                case F.GlyphClass:
                    classes.append(data)
                case F.AxisCount:
                    font._axis_count = data
                case F.AxisName:
                    font.fake_deserialize_axis(data)
                case F.AnisotropicInterpolationMappings:
                    font._anisotropic_interpolation_mappings = data
                case F.AxisMappingsCount:
                    font._axis_mappings_count = data
                case F.AxisMappings:
                    font._axis_mappings = data
                case M.MasterName:
                    font._master_names.append(data)
                case M.MasterLocation:
                    font._master_locations.append(data)
                case F.PrimaryInstanceLocations:
                    font._primary_instance_locations = data
                case F.PrimaryInstances:
                    font._primary_instances = data
                case M.PostScriptInfo:
                    font._master_ps_infos.append(data)
                case F.GlobalGuides:
                    font.fake_deserialize_guides(data)
                case F.GlobalGuideProperties:
                    font.fake_deserialize_guide_properties(data)
                case F.GlobalMask:
                    pass
                case G.Glyph:
                    # Append the current glyph
                    if glyph is not None:
                        font.glyphs.append(glyph)
                    # Make a new glyph
                    logger.info(f"Adding Glyph: '{data.get('name')}', {key}")
                    glyph = Glyph()
                    # Add the data
                    glyph.fake_deserialize(2001, data)
                case F.OpenTypeExportOptions:
                    font._ot_export_options = data
                case F.ExportOptions:
                    font._export_options = data
                case F.MappingMode:
                    font._mapping_mode = data
                case F.E1410:
                    pass
                case _:
                    logger.error(f"Unhandled VFB entry: {key}")

        if glyph is not None:
            font.glyphs.append(glyph)

        if gids:
            enc = font._encoding = Encoding()
            enc._parent = font

            max_gid = min(255, max(gids))
            for i in range(max_gid + 1):
                er = EncodingRecord()
                if i in gids:
                    er.name = gids[i]
                    er.unicode = min(self.nametable.get_unicodes_for_name(er.name))
                else:
                    er.name = f"_{i:04d}"
                enc.append(er)

        if classes:
            # The setter can't append, assign the whole list
            font._classes.fake_set_classes(classes)



from __future__ import annotations

import argparse
import logging
from code import InteractiveConsole
from pathlib import Path
from typing import Any

from vfbLib.json import save_vfb_json

from FL import environment, fl

try:
    from ptpython.repl import embed

    have_ptpython = True
except ImportError:
    have_ptpython = False

try:
    from IPython import embed as iembed

    have_ipython = True
except ImportError:
    have_ipython = False

logger = logging.getLogger(__name__)


__startup_code__ = """
from FL import fl
"""


class FontLab5Console(InteractiveConsole):
    def __init__(
        self, startup_code: str = "", locals: dict[str, Any] | None = None
    ) -> None:
        namespace = locals or {}
        # code = compile(__startup_code__, "", "exec", 0)
        # exec(code, namespace)
        super().__init__(locals=namespace)


def main() -> None:
    parser = argparse.ArgumentParser(description="FontLab 5 external scripting")
    parser.add_argument(
        "-o",
        "--out-path",
        type=str,
        nargs=1,
        help="Save files to output path instead of overwriting the original files",
    )
    parser.add_argument(
        "-d",
        "--no-decompile",
        action="store_true",
        default=False,
        help="When roundtripping, don't decompile entries in JSON file",
    )
    parser.add_argument(
        "-r",
        "--roundtrip",
        action="store_true",
        default=False,
        help="Roundtrip specified VFB file(s) through FakeLab, then exit",
    )
    parser.add_argument(
        "-s",
        "--script",
        type=str,
        nargs="+",
        help="Path(s) to Python scripts to run on the VFB file(s)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_argument(
        "vfb",
        type=str,
        nargs="*",
        help=(
            "Path(s) to VFB file(s) to operate on. "
            "The last file will be the current font"
        ),
    )
    args = parser.parse_args()
    if args:
        if args.verbose:
            logging.basicConfig(level=logging.INFO)
        if args.roundtrip:
            for vfb_path in args.vfb:
                logger.info(vfb_path)
                fl.Open(vfb_path, addtolist=False)
                out_path = Path(vfb_path).with_suffix(".fake.vfb")
                fl.Save(str(out_path))
                save_vfb_json(out_path, no_decompile=args.no_decompile)
        else:
            for vfb_path in args.vfb:
                logger.info(vfb_path)
                fl.Open(vfb_path, addtolist=True)
            if args.script:
                # If we have scripts, run them and exit.
                for script_path in args.script:
                    exec(Path(script_path).read_text(), locals=environment)
            else:
                # Run the interactive console.
                if have_ptpython:
                    # globals=globals()?
                    embed(locals=environment, vi_mode=True)
                elif have_ipython:
                    iembed(globals=globals(), locals=environment, header="FakeLab")
                else:
                    console = FontLab5Console(locals=environment)
                    console.interact(
                        banner="Welcome to the Headless FakeLab REPL.",
                        exitmsg="Happy fonting, my friend!",
                    )
    else:
        parser.print_help()



import unittest

import pytest

from FL.objects.Matrix import Matrix
from FL.objects.Point import Point


class PointTests(unittest.TestCase):
    def test_instantiation(self):
        p = Point()
        assert p.x == 0.0
        assert p.y == 0.0

    def test_instantiation_coords(self):
        p = Point(-200, 0)
        assert p.x == -200.0
        assert p.y == 0.0

    def test_instantiation_coords_invalid(self):
        with pytest.raises(RuntimeError):
            Point(-200)

    def test_instantiation_copy(self):
        p = Point(-200, 0)
        p2 = Point(p)
        assert p.x == -200.0
        assert p.y == 0.0
        assert p2 == p
        p.x += 1
        assert p.x == -199.0
        # Make sure the points are independent objects
        assert p != p2

    def test_instantiation_copy_invalid(self):
        p = Point()
        with pytest.raises(RuntimeError):
            Point(p, 5)

    def test_addition(self):
        p = Point(-199, 0)
        p2 = Point(-200, 0)
        p2.Add(p)
        assert p2.x == -399.0
        assert p2.y == 0.0

    # Multiplication

    def test_mul_Point(self):
        p = Point(44, 55)
        p2 = Point(2, 3)
        assert p * p2 == 253.0

    def test_mul_Matrix(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        p = Point(44, 55)
        p *= m1
        assert (p.x, p.y) == (340, 1296)

    def test_mul_float(self):
        p = Point(44, 55)
        p *= 2
        assert (p.x, p.y) == (88, 110)

    # Transform

    def test_Transform_Point(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        p = Point(44, 55)
        p.Transform(m1)
        assert (p.x, p.y) == (340, 1296)

    def test_Transform_Point_float(self):
        m1 = Matrix(2, 4, 8, 16, 32, 64)
        p = Point(44.2, 55.5)
        p.Transform(m1)
        assert (p.x, p.y) == (342.4, 1305.6)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from FL.objects.Point import Point
    from FL.objects.Rect import Rect


class Dialog:
    """
    Dialog - class to represent Dialog

    Dialog.__doc__ reports only 'Dialog' ... so use this information with care)

    See 'yourFontLabFolder/Macros/Effects/digital.py' for an example
    """

    # Constructor

    def __init__(self) -> None:
        """
        Dialog(DialogClass) - generic constructor
        """
        pass

    # Attributes

    @property
    def size(self) -> Point:
        raise NotImplementedError

    @size.setter
    def size(self, value: Point) -> None:
        """
        Size of the dialog as a Point object

        Args:
            value (Point): _description_
        """
        raise NotImplementedError

    @property
    def title(self) -> str:
        raise NotImplementedError

    @title.setter
    def title(self, value: str) -> None:
        """
        Title of the dialog

        Args:
            value (str): _description_
        """
        raise NotImplementedError

    # Methods

    def Center(self) -> None:
        """
        Center dialog on screen
        """
        raise NotImplementedError

    def AddControl(
        self, ControlType: int, dimension: Rect, identifier: str, ControlStyle: int
    ) -> None:
        """
        Adds a control to the dialog, see `FLdict` for Type- and Style-constants

        Args:
            ControlType (int): _description_
            dimension (Rect): _description_
            identifier (str): _description_
            ControlStyle (int): _description_
        """
        raise NotImplementedError

    def GetValue(self, identifier: str) -> Any:
        """
        Get value from the specified control

        Args:
            identifier (str): _description_

        Returns:
            Any: _description_
        """
        raise NotImplementedError

    def PutValue(self, identifier: str) -> None:
        """
        Put value to the specified control

        Args:
            identifier (str): _description_
        """
        raise NotImplementedError

    def Enable(self, identifier: str, value: int) -> None:
        """
        Enable (value=1) or disable (value=0) the specified control

        Args:
            identifier (str): _description_
            value (int): _description_
        """
        raise NotImplementedError

    def Show(self, identifier: str, value: int) -> None:
        """
        Show (value=1) or hide (value=0) the specified control

        Args:
            identifier (str): _description_
            value (int): _description_
        """
        raise NotImplementedError

    def SetLabel(self, identifier: str, value: str) -> None:
        """
        Set the label of the specified control

        Args:
            identifier (str): _description_
            value (str): _description_
        """
        raise NotImplementedError

    def GetRect(self, identifier: str) -> Rect:
        """
        Get dimension of the specified control as a Rect object

        Args:
            identifier (str): _description_
        """
        raise NotImplementedError

    def Repaint(self, identifier: str) -> None:
        """
        Redraw the specified (custom)control

        Args:
            identifier (str): _description_
        """
        raise NotImplementedError

    def Run(self) -> int:
        """
        Run the dialog

        Returns:
            int: _description_
        """
        # Return type is deduced from digital.py
        raise NotImplementedError

    def End(self, returnvalue: int | None = None):
        """
        End the dialog

        Args:
            returnvalue (int | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError



from __future__ import annotations

import logging

from vfbLib.typing import GuidePropertiesDict, GuidePropertyDict, MMGuidesDict

from FL.objects.Guide import Guide

logger = logging.getLogger(__name__)


class GuideMixin:
    def fake_deserialize_guides(self, data: MMGuidesDict) -> None:
        for k, target in (("h", self.hguides), ("v", self.vguides)):
            if dir_master_guides := data[k]:
                for guide_dict in dir_master_guides:
                    g = Guide(guide_dict[0]["pos"], guide_dict[0]["angle"])
                    for master_index in range(1, self._masters_count):
                        g._positions[master_index] = guide_dict[master_index]["pos"]
                        g._widths[master_index] = Guide.fake_angle_to_width(
                            guide_dict[master_index]["angle"]
                        )
                    target.append(g)

    def fake_serialize_guides(self) -> MMGuidesDict:
        mgd = MMGuidesDict(h=[], v=[])
        for direction, source in (("h", self.hguides), ("v", self.vguides)):
            for guide in source:
                guide_masters = []
                for master_index in range(self._masters_count):
                    guide_masters.append(
                        {
                            "pos": guide.positions[master_index],
                            "angle": Guide.fake_width_to_angle(
                                guide.widths[master_index]
                            ),
                        }
                    )
                mgd[direction].append(guide_masters)
        return mgd


class GuidePropertiesMixin:
    def fake_deserialize_guide_properties(self, data: GuidePropertiesDict) -> None:
        for k, target in (("h", self.hguides), ("v", self.vguides)):
            num_guides = len(target)
            for guide_prop_dict in data[k]:
                guide_index = guide_prop_dict["index"] - 1
                if guide_index >= num_guides:
                    logger.info(
                        "Skipping properties for guide that isn't present: "
                        f"{guide_index} in {target}"
                    )
                    continue

                color = guide_prop_dict.get("color")
                if color:
                    target[guide_index]._color = color
                name = guide_prop_dict.get("name")
                if name:
                    target[guide_index]._name = name

    def fake_serialize_guide_properties(self) -> GuidePropertiesDict:
        gpl = GuidePropertiesDict(h=[], v=[])
        for k, target in (("h", self.hguides), ("v", self.vguides)):
            for i, guide in enumerate(target):
                d = GuidePropertyDict(index=i + 1)
                if guide._color:
                    d["color"] = guide._color
                if guide._name:
                    d["name"] = guide._name
                if "color" in d or "name" in d:
                    gpl[k].append(d)
        return gpl



import unittest
from pathlib import Path

import pytest

from FL.helpers.registry import parse_registry_file


def get_reg_path(filename):
    return Path(__file__).parent.parent / "data" / filename


class RegistryTests(unittest.TestCase):
    def test_parse_ok(self):
        reg = parse_registry_file(get_reg_path("minimal.reg"))
        assert reg == {
            "CacheTTPath": "",
            "ExpandKernCodepage": "MS Windows 1252 Western (ANSI)",
            "ExpandKernCount": 10919,
        }

    def test_parse_empty(self):
        with pytest.raises(EOFError):
            parse_registry_file(get_reg_path("empty.reg"))

    def test_parse_no_key(self):
        with pytest.raises(ValueError):
            parse_registry_file(get_reg_path("nokey.reg"))

    def test_parse_wrong_header(self):
        with pytest.raises(AssertionError):
            parse_registry_file(get_reg_path("wrong_header.reg"))

    def test_parse_wrong_value_type(self):
        with pytest.raises(AssertionError):
            parse_registry_file(get_reg_path("wrong_value_type.reg"))



from __future__ import annotations

import array
from typing import TYPE_CHECKING

from vfbLib.typing import BackgroundImageDict, BitmapDataDict

from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph
    from FL.objects.Rect import Rect


class Image:
    __slots__ = [
        "_data",
        "_empty",
        "_size",
        "_traceenabled",
        "_height",
        "_width",
        # Non-API
        "_flag",
        "_origin",
        "_preview",
        "_size_units",
    ]

    def __init__(
        self, width_or_image: int | Image | None = None, height: int | None = None
    ) -> None:
        """
        Image - class to represent image

        Image() - generic constructor, creates an Image with zero coordinates
        Image(Image) - copy constructor
        Image(int width, int height) - creates an Image of given size
        """
        self._empty = True
        self._data = array.array("H")
        self._size = len(self._data)
        self._width = 0
        self._height = 0
        self._traceenabled = True

        # Non-API
        self._flag = 0
        self._origin = Point()
        self._size_units = Point()

        if isinstance(width_or_image, int):
            self._width = width_or_image
            if not isinstance(height, int):
                raise TypeError
            self._height = height
            self._empty = True
        elif isinstance(width_or_image, Image):
            # copy constructor
            image = width_or_image
            self._width = image.width
            self._height = image.height
            self._empty = bool(image.empty)
            self._traceenabled = bool(image.traceenabled)
            self._data = array.array("H", image.data)
            self._size = image.size
        # else empty image

    # Additions for FakeLab

    def fake_deserialize(self, data: BackgroundImageDict) -> None:
        x, y = data["origin"]
        self._origin.x = x
        self._origin.y = y
        x, y = data["size_units"]
        self._size_units.x = x
        self._size_units.y = y
        x, y = data["size_pixels"]
        self._width = x
        self._height = y
        self._flag = data["bitmap"]["flag"]
        self._data = array.array("H", data["bitmap"]["data"])
        self._empty = False

    def fake_serialize(self) -> BackgroundImageDict | None:
        if self._empty:
            return None

        d = BackgroundImageDict(
            origin=(
                int(self._origin.x),
                int(self._origin.y),
            ),
            size_units=(
                int(self._size_units.x),
                int(self._size_units.y),
            ),
            size_pixels=(self.width, self.height),
            bitmap=BitmapDataDict(flag=self._flag, data=self.data),
        )
        return d

    # Attributes

    @property
    def width(self) -> int:
        """
        Dimensions of the image

        Returns:
            int: The width of the image in pixels
        """
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        raise RuntimeError("Class Image does not have writable attributes")

    @property
    def height(self) -> int:
        """
        Dimensions of the image

        Returns:
            int: The height of the image in pixels
        """
        return self._height

    @height.setter
    def height(self, value: int) -> None:
        raise RuntimeError("Class Image does not have writable attributes")

    @property
    def empty(self) -> int:
        """
        The empty status of the image

        Returns:
            int: Whether the image is empty
        """
        return int(self._empty)

    @empty.setter
    def empty(self, value: int) -> None:
        raise RuntimeError("Class Image does not have writable attributes")

    @property
    def size(self) -> int:
        """
        The size of the image buffer

        Returns:
            int: The size of the image buffer
        """
        return len(self.data)

    @size.setter
    def size(self, value: int) -> None:
        raise RuntimeError("Class Image does not have writable attributes")

    @property
    def data(self) -> list[int]:
        """
        Access the image buffer

        Returns:
            list[int] | None: _description_
        """
        return [i for i in self._data]

    @data.setter
    def data(self, value: bytes | None) -> None:
        # if value is None:
        #     self._data = array.array("H", [0] * self.width * self.height)
        # else:
        #     self._data = array.array("H", value)
        # Does nothing
        pass

    @property
    def traceenabled(self) -> int:
        """
        There is a possibility to trace the image with the Trace command

        Returns:
            int: Whether the image can be traced
        """
        return int(self._traceenabled)

    @traceenabled.setter
    def traceenabled(self, value: int) -> None:
        raise RuntimeError("Class Image does not have writable attributes")

    # Methods

    def Create(self, width: int, height: int) -> None:
        """
        Creates a blank image of given size.

        Args:
            width (int): The width in pixels
            height (int): The height in pixels
        """
        raise NotImplementedError

    def Trace(self, glyph: Glyph) -> None:
        """
        Traces the image with the current option and adds it to the glyph

        Args:
            glyph (Glyph): The glyph to which the traced outline will be added
        """
        raise NotImplementedError

    def Clear(self) -> None:
        """
        Clears the image
        """
        # (not reported by docstring)
        raise NotImplementedError

    def GetPixel(self, p: Point) -> int:
        # (not reported by docstring)
        raise NotImplementedError

    def HLine(self, x0: int, x1: int, y: int, color: int) -> None:
        # (not reported by docstring)
        raise NotImplementedError

    def ImageBlt(self, dest: Image, source_rect: Rect, dest_point: Point) -> None:
        # (not reported by docstring)
        raise NotImplementedError

    def Invert(self) -> None:
        # (not reported by docstring)
        raise NotImplementedError

    def SetPixel(self, p: Point) -> None:
        # (not reported by docstring)
        raise NotImplementedError



from __future__ import annotations

from collections import UserList
from copy import copy
from typing import Any, Iterable, TypeVar

T = TypeVar("T")


class ListParent(UserList[T]):
    """
    Like a list, but set the _parent attribute for each item.
    """

    def __init__(self, iterable: Iterable[T] = [], parent: Any | None = None) -> None:
        super().__init__(copy(iterable))
        self._parent = parent

    def __add__(self, item: Any) -> ListParent:
        # Makes scripting unresponsive in FL5
        # We raise an error that is not used otherwise
        raise ReferenceError

    def __iadd__(self, item: Any) -> ListParent:
        # Makes scripting unresponsive in FL5
        # We raise an error that is not used otherwise
        raise ReferenceError

    # def __radd__(self, item: Any) -> None:
    #     item._parent = self._parent
    #     self.data.__radd__(item)

    def __setitem__(self, index: int, item: Any) -> None:
        item._parent = self._parent
        self.data[index] = item

    def append(self, item: Any) -> None:
        item._parent = self._parent
        self.data.append(item)

    def clear(self) -> None:
        # Not implemented, probably a mix-up with clean()?
        # Raise AttributeError as in FL5
        raise AttributeError

    def extend(self, iterable: Iterable) -> None:
        # Not implemented
        # Raise AttributeError as in FL5
        raise AttributeError

    def insert(self, index: int, item: Any) -> None:
        item._parent = self._parent
        self.data.insert(index, item)

    # FontLab-specific

    def clean(self) -> None:
        """
        Remove all items from the list.
        """
        self.data = []



import unittest

import pytest

from FL.objects.TTHCommand import TTHCommand


class TTHCommandTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        # No args
        with pytest.raises(RuntimeError):
            TTHCommand()

    def test_instantiation_args(self) -> None:
        # Initialize with code and one param
        t = TTHCommand(1, 1, 2)
        assert t.code == 1
        assert t.params == [1, 2]
        assert str(t) == "<TTHCommand: ALIGNTOP(1, 2)>"

    def test_instantiation_args_too_many(self) -> None:
        # Initialize with code and too many params
        # The message says one command and up to 5 integer params, but the error is
        # thrown at 1 + 4 params already.
        with pytest.raises(RuntimeError):
            TTHCommand(1, 1, 2, 3, 4, 5)



from __future__ import annotations

# from fontTools.misc.textTools import deHexStr, hexStr
from typing import TYPE_CHECKING, Any

from vfbLib.enum import F, G, M, T
from vfbLib.vfb.entry import VfbEntry
from vfbLib.vfb.header import VfbHeader
from vfbLib.vfb.vfb import Vfb

from FL.objects.Font import Font
from FL.objects.TTInfo import TTInfo

if TYPE_CHECKING:
    from enum import IntEnum
    from pathlib import Path


class FontToVfbWriter:
    """
    Convert the Font object to the low-level `vfbLib.vfb.vfb.Vfb` structure and write it
    to `vfb_path`
    """

    def __init__(self, font: Font, vfb_path: Path) -> None:
        """Save the `font` into the VFB file at `vfb_path`.

        Args:
            font (Font): The source object of the data
            vfb_path (Path): The file path to which to write the VFB data
        """
        self.vfb_path = vfb_path
        self.font = font
        self.vfb = Vfb()
        self.compile()
        self.vfb.write(self.vfb_path)

    def add_direct_entries(
        self, keys: tuple[int, ...], parent: Font | TTInfo, enum: type[IntEnum] = F
    ) -> None:
        for key in keys:
            attr = enum(key).name
            self.add_entry(key, getattr(parent, attr))

    def add_entry(self, eid: int, data: Any) -> None:
        e = VfbEntry(self.vfb, eid=eid)
        e.data = data
        self.vfb.entries.append(e)

    def compile(self) -> None:
        self.compile_header()
        self.compile_encoding()
        self.compile_font_info()
        self.compile_glyphs()
        self.compile_options()
        # That's all, folks

    def compile_header(self) -> None:
        header = self.vfb.header = VfbHeader()
        # fmt: off
        header.data = {
            "header0": 26,
            "filetype": "WLF10",
            "header1": 3,
            "chunk1": [
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 1, 0, 0, 0, 4, 0, 0, 0, 10, 0
            ],
            "creator": {1: 1, 2: [5, 2, 2, 128], 3: 0},
            "end0": 6,
            "end1": 1,
            "end2": 0,
        }
        # fmt: on

    def compile_encoding(self) -> None:
        for i in range(len(self.font._encoding_default)):
            self.add_entry(F.EncodingDefault, [i, self.font._encoding_default[i].name])
        for i in range(len(self.font.encoding)):
            self.add_entry(F.Encoding, [i, self.font.encoding[i].name])

        # We don't know what these do exactly:
        # Sometimes 0, sometimes 1, sometimes 42694?
        self.add_entry(F.E1502, self.font._unknown_pleasures[1502])
        # Apparently always empty:
        self.add_entry(F.E518, self.font._unknown_pleasures[518])
        self.add_entry(F.E257, self.font._unknown_pleasures[257])

    def compile_font_info(self) -> None:
        font = self.font
        self.vfb.num_masters = font._masters_count
        self.add_entry(F.font_name, font.font_name)
        self.add_entry(F.MasterCount, self.vfb.num_masters)
        self.add_entry(F.weight_vector, font.weight_vector[: self.vfb.num_masters])
        self.add_direct_entries(
            (
                F.unique_id,
                F.version,
                F.notice,
                F.full_name,
                F.family_name,
                F.pref_family_name,
                F.menu_name,
                F.apple_name,
                F.weight,
                F.width,
            ),
            font,
        )
        self.add_entry(F.License, font._license)
        self.add_entry(F.LicenseURL, font._license_url)
        self.add_direct_entries(
            (
                F.copyright,
                F.trademark,
                F.designer,
                F.designer_url,
                F.vendor_url,
                F.source,
                F.is_fixed_pitch,
                F.weight_code,
                F.italic_angle,
                F.slant_angle,
                F.underline_position,
                F.underline_thickness,
                F.ms_charset,
                F.panose,
                F.tt_version,
                F.tt_u_id,
                F.style_name,
                F.pref_style_name,
                F.mac_compatible,
            ),
            font,
        )
        self.add_entry(1140, font._unknown_pleasures[1140])
        self.add_direct_entries(
            (
                F.vendor,
                F.xuid,
                F.xuid_num,
                F.year,
                F.version_major,
                F.version_minor,
                F.upm,
                F.fond_id,
            ),
            font,
        )
        self.add_entry(F.PostScriptHintingOptions, font._postscript_hinting_options)
        self.add_entry(1068, font._unknown_pleasures[1068])
        self.add_direct_entries(
            (
                F.blue_values_num,
                F.other_blues_num,
                F.family_blues_num,
                F.family_other_blues_num,
                F.stem_snap_h_num,
                F.stem_snap_v_num,
                F.font_style,
            ),
            font,
        )
        if font.pcl_id >= 0:
            self.add_entry(F.pcl_id, font.pcl_id)
        if font.vp_id >= 0:
            self.add_entry(F.vp_id, font.vp_id)

        self.add_direct_entries(
            (
                F.ms_id,
                F.pcl_chars_set,
            ),
            font,
        )

        self.compile_ttinfo()

        # Special handling required:
        self.add_entry(F.fontnames, [nr.fake_serialize() for nr in font.fontnames])
        self.add_entry(F.CustomCMAPs, font._custom_cmaps)
        self.add_entry(F.PCLTTable, font._pclt_table)
        self.add_entry(F.ExportPCLTTable, font._export_pclt_table)
        if font.note:
            self.add_entry(F.note, font.note)
        self.add_entry(2030, font._unknown_pleasures[2030])
        if font.customdata:
            self.add_entry(F.customdata, font.customdata)

        for ttt in font.truetypetables:
            self.add_entry(F.TrueTypeTable, ttt)

        if metrics_class_flags := font._metrics_class_flags:
            self.add_entry(F.MetricsClassFlags, metrics_class_flags)
        if kerning_class_flags := font._kerning_class_flags:
            self.add_entry(F.KerningClassFlags, kerning_class_flags)

        fea = font.fake_serialize_features()
        if fea:
            self.add_entry(F.features, fea)

        for ot_class in font._classes:
            self.add_entry(F.GlyphClass, ot_class)

        self.add_entry(513, font._unknown_pleasures[513])
        self.add_entry(271, font._unknown_pleasures[271])

        self.add_entry(F.AxisCount, len(font.axis))

        for axis_name in font.fake_serialize_axis():
            self.add_entry(F.AxisName, axis_name)

        self.add_entry(
            F.AnisotropicInterpolationMappings, font._anisotropic_interpolation_mappings
        )
        self.add_entry(F.AxisMappingsCount, font._axis_mappings_count)
        self.add_entry(F.AxisMappings, font._axis_mappings)

        for master_index in range(self.vfb.num_masters):
            self.add_entry(M.MasterName, font._master_names[master_index])
            self.add_entry(M.MasterLocation, font._master_locations[master_index])

        if font._primary_instance_locations:
            self.add_entry(F.PrimaryInstanceLocations, font._primary_instance_locations)
        if font._primary_instances:
            self.add_entry(F.PrimaryInstances, font._primary_instances)

        for master_ps_info in font._master_ps_infos:
            self.add_entry(M.PostScriptInfo, master_ps_info)

        self.add_entry(527, font._unknown_pleasures[527])

        if font.hguides or font.vguides:
            self.add_entry(F.GlobalGuides, font.fake_serialize_guides())
            self.add_entry(
                F.GlobalGuideProperties, font.fake_serialize_guide_properties()
            )

        if font.default_character:
            self.add_entry(F.default_character, font.default_character)

    def compile_ttinfo(self) -> None:
        for k in (T.cvt, T.prep, T.fpgm):
            d = self.font.ttinfo.fake_get_binary(T(k).name)
            if d:
                self.add_entry(k, d)

        if gasp := self.font.ttinfo.fake_serialize_gasp():
            self.add_entry(T.gasp, gasp)

        self.add_entry(F.ttinfo, self.font.ttinfo.fake_serialize())
        self.add_entry(T.vdmx, self.font.ttinfo.fake_serialize_vdmx())
        self.add_direct_entries(
            (T.hhea_line_gap, T.hhea_ascender, T.hhea_descender), self.font.ttinfo, T
        )
        # "TrueType Stem PPEMS 2 And 3" are skipped, it is apparently a remnant of some
        # old FontLab version. The data is also contained in "TrueType Stem PPEMs"
        self.add_entry(
            T.TrueTypeStemPPEMs, self.font.ttinfo.fake_serialize_stem_ppems()
        )
        self.add_entry(T.TrueTypeStems, self.font.ttinfo.fake_serialize_stems())
        self.add_entry(
            T.TrueTypeStemPPEMs1, self.font.ttinfo.fake_serialize_stem_ppems1()
        )
        self.add_entry(T.TrueTypeZones, self.font.ttinfo.fake_serialize_zones())
        self.add_entry(F.unicoderanges, self.font.unicoderanges)  # Not TTInfo
        self.add_entry(T.stemsnaplimit, self.font.ttinfo._stemsnaplimit)
        self.add_entry(T.zoneppm, self.font.ttinfo._zoneppm)
        self.add_entry(T.codeppm, self.font.ttinfo._codeppm)

        self.add_entry(1604, self.font.ttinfo._unknown_pleasures[1604])
        self.add_entry(2032, self.font.ttinfo._unknown_pleasures[2032])

        self.add_entry(
            T.TrueTypeZoneDeltas, self.font.ttinfo.fake_serialize_zone_deltas()
        )

    def compile_glyphs(self) -> None:
        for glyph in self.font.glyphs:
            glyph_dict = glyph.fake_serialize()
            # TODO: Which keys are required?
            for key in (
                G.Glyph,
                G.Links,
                G.image,  # FIXME
                G.Bitmaps,  # FIXME
                G.E2023,
                G.Sketch,  # FIXME
                G.HintingOptions,
                G.mask,
                G.MaskMetrics,
                G.MaskMetricsMM,
                G.Origin,
                G.unicodes,
                G.E2034,
                G.UnicodesNonBMP,
                G.mark,
                G.customdata,
                G.note,
                G.GDEFData,
                G.AnchorsProperties,
                G.AnchorsMM,
                G.GuideProperties,
            ):
                if key in glyph_dict:
                    self.add_entry(key, glyph_dict[key])

    def compile_options(self) -> None:
        if ot_export_options := self.font._ot_export_options:
            self.add_entry(F.OpenTypeExportOptions, ot_export_options)
        if export_options := self.font._export_options:
            self.add_entry(F.ExportOptions, export_options)

        self.add_entry(F.MappingMode, self.font._mapping_mode)



from __future__ import annotations

from typing import Any


class Copyable:
    __slots__: list[str] = ["_parent"]

    def _copy_constructor(self, other: Any) -> None:
        from copy import deepcopy

        assert isinstance(other, type(self))

        tmp = deepcopy(other)
        for attr in self.__slots__:
            if attr == "_parent":
                # Parent is not copied
                self._parent: Any | None = None
            else:
                setattr(self, attr, getattr(tmp, attr))



from __future__ import annotations


class TTPoint:
    """
    TTPoint - class to represent TrueType points

    All coordinates in TTPoint class are set in 10.6 fixed-point format
    """

    __slots__ = ["_x", "_y", "_flag"]

    # Constructor

    def __init__(self) -> None:
        """
        TTPoint() - generic constructor, creates a TTPoint with zero coordinates
        """
        # "All coordinates in TTPoint class are set in 10.6 fixed-point format"
        # ... so 32767 is 0!?
        self.x = 32767
        self.y = 32767
        self.flag = 255

    def __repr__(self) -> str:
        return f"<TTPoint: flags: {self.flag}, x: {self.x} y: {self.y}, Orphan>"

    # Attributes

    @property
    def x(self) -> int:
        """
        x coordinate

        Returns:
            int: _description_
        """
        return self._x

    @x.setter
    def x(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        self._x = value

    @property
    def y(self) -> int:
        """
        y coordinate

        Returns:
            int: _description_
        """
        return self._y

    @y.setter
    def y(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        self._y = value

    @property
    def flag(self) -> int:
        """
        point's flag (on-curve or off-curve)

        Returns:
            int: _description_
        """
        return self._flag

    @flag.setter
    def flag(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        self._flag = value

    # Operations: none

    # Methods: none



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable
from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Matrix import Matrix


class Rect(Copyable):
    """
    Rect()                   - generic constructor, creates a Rect with zero coordinates
    Rect(Rect r)             - copy constructor
    Rect(Point p)            - creates rectangle with one corner at (0, 0) and another
                               at coordinates defined by p
    Rect(Point p0, Point p1) - creates rectangle defined by the corner points
    Rect(x0, y0, x1, y1)     - creates a rectangle defined by the coordinates of the
                               corner points
    """

    __slots__ = ["_x0", "_y0", "_x1", "_y1"]

    # Constructor

    def __init__(
        self,
        r_or_p0_or_x0: Rect | Point | float | None = None,
        p1_or_x0: Point | float | None = None,
        x1: float | None = None,
        y1: float | None = None,
    ) -> None:
        self._x0 = 0.0
        self._y0 = 0.0
        self._x1 = 0.0
        self._y1 = 0.0

        if r_or_p0_or_x0 is not None:
            self.Assign(r_or_p0_or_x0, p1_or_x0, x1, y1)

    def __repr__(self) -> str:
        return "<Rect: %g,%g,%g,%g>" % (self._x0, self._y0, self._x1, self._y1)

    # Attributes

    @property
    def ll(self) -> Point:
        """
        position of the left/bottom corner
        """
        return Point(self._x0, self._y0)

    @property
    def ur(self) -> Point:
        """
        position of the right/top corner
        """
        return Point(self._x1, self._y1)

    @property
    def x(self) -> float:
        return self._x0

    @x.setter
    def x(self, value: float) -> None:
        self._x0 = float(value)

    @property
    def y(self) -> float:
        return self._y0

    @y.setter
    def y(self, value: float) -> None:
        self._y0 = float(value)

    @property
    def width(self) -> float:
        return self._x1 - self._x0

    @width.setter
    def width(self, value: float) -> None:
        self._x1 = self._x0 + value

    @property
    def height(self) -> float:
        return self._y1 - self._y0

    @height.setter
    def height(self, value: float) -> None:
        self._y1 = self._y0 + value

    # Operations

    def __add__(self, other: Point | Rect) -> Rect:
        """
        Point or Rect must be second operand, rectangle is expanded to include this
        point or rectangle

        Args:
            other (Point | Rect): The point or rectangle to add.

        Returns:
            Rect: The enclosing rectangle.
        """
        # This again doesn't check that the corner points are normalized, so we can
        # start with a "negative" rect spanning the full area, and have easier
        # comparisons.
        if isinstance(other, Point):
            if other.x < self._x0:
                self._x0 = other.x
            if other.y < self._y0:
                self._y0 = other.y
            if other.x > self._x1:
                self._x1 = other.x
            if other.y > self._y1:
                self._y1 = other.y
        elif isinstance(other, Rect):
            if other.ll.x < self._x0:
                self._x0 = other.ll.x
            if other.ll.y < self._y0:
                self._y0 = other.ll.y
            if other.ur.x > self._x1:
                self._x1 = other.ur.x
            if other.ur.y > self._y1:
                self._y1 = other.ur.y
        else:
            raise TypeError
        return self

    def __mul__(self, other: Matrix) -> Rect:
        """
        Second operand must be `Matrix`. Matrix transformation is applied to the
        rectangle.
        """
        raise NotImplementedError

    # Methods

    def Assign(
        self,
        r_or_p0_or_x0: Rect | Point | float | None = None,
        p1_or_x0: Point | float | None = None,
        x1: float | None = None,
        y1: float | None = None,
    ) -> None:
        """
        Assigns new values to a Rect, the same as constructor

        Args:
            r_or_p0_or_x0 (Rect | Point | float | None, optional): _description_. Defaults to None.
            p1_or_x0 (Point | float | None, optional): _description_. Defaults to None.
            x1 (float | None, optional): _description_. Defaults to None.
            y1 (float | None, optional): _description_. Defaults to None.
        """
        if isinstance(r_or_p0_or_x0, Rect):
            # Copy constructor
            self._copy_constructor(r_or_p0_or_x0)
        elif isinstance(r_or_p0_or_x0, Point):
            if isinstance(p1_or_x0, Point):
                self._x0 = r_or_p0_or_x0.x
                self._y0 = r_or_p0_or_x0.y
                self._x1 = p1_or_x0.x
                self._y1 = p1_or_x0.y
            else:
                # The doc is not correct, x/y don't stay at 0,
                # but the point is used for all coords
                self._x0 = r_or_p0_or_x0.x
                self._y0 = r_or_p0_or_x0.y
                self._x1 = r_or_p0_or_x0.x
                self._y1 = r_or_p0_or_x0.y
        elif isinstance(r_or_p0_or_x0, int) or isinstance(r_or_p0_or_x0, float):
            assert isinstance(p1_or_x0, int) or isinstance(p1_or_x0, float)
            assert x1 is not None
            assert y1 is not None
            self._x0 = float(r_or_p0_or_x0)
            self._y0 = float(p1_or_x0)
            self._x1 = float(x1)
            self._y1 = float(y1)

    def Shift(self, p_or_x: Point | float, y: float | None = None) -> None:
        # (Point p) | (x, y) - shifts Rect on a position defined by p or x
        # and y values
        raise NotImplementedError

    def Transform(self, m: Matrix) -> None:
        # (Matrix m) - applies Matrix transformation to the Rect (see Matrix().__doc__)
        raise NotImplementedError

    def Resize(self) -> None:
        # (width, height) - resizes rectangle to new width and height
        raise NotImplementedError

    def Include(
        self, r_or_p_or_x: Rect | Point | float, y: float | None = None
    ) -> None:
        # Expands rectangle to include new rectangle or point
        if isinstance(r_or_p_or_x, Point):
            # This again doesn't check that the corner points are normalized.
            if r_or_p_or_x.x < self._x0:
                self._x0 = r_or_p_or_x.x
            if r_or_p_or_x.y < self._y0:
                self._y0 = r_or_p_or_x.y
            if r_or_p_or_x.x > self._x1:
                self._x1 = r_or_p_or_x.x
            if r_or_p_or_x.y > self._y1:
                self._y1 = r_or_p_or_x.y
        elif isinstance(r_or_p_or_x, Rect):
            raise NotImplementedError
        else:
            # floats
            if not isinstance(y, float):
                raise RuntimeError

            raise NotImplementedError

    def Check(self, r_or_p: Rect | Point) -> bool:
        # (Rect r) - returns True if r overlaps current rectangle
        # (Point p) - returns True if p is inside current rectangle
        raise NotImplementedError

    def Validate(self) -> None:
        """
        Corrects rectangle's orientation
        """
        self._x0, self._x1 = sorted([self._x0, self._x1])
        self._y0, self._y1 = sorted([self._y0, self._y1])



from __future__ import annotations

from typing import Any

from FL.fake.Base import Copyable
from FL.objects.Matrix import Matrix


class Point(Copyable):
    __slots__ = ["_parent", "_x", "_y"]

    # Constructor

    def __init__(
        self, p_or_x: Point | float | None = None, y: float | None = None
    ) -> None:
        """
        Point - base class to represent point

        Args:
            p_or_x (Point | float | None, optional): The `Point` to copy coordinates
                from, or the x coordinate. Defaults to None.
            y (float | None, optional): The y coordinate. Defaults to None.
        """
        self._parent = None
        self.x = 0
        self.y = 0
        if p_or_x is not None:
            self.Assign(p_or_x, y)

    def __repr__(self) -> str:
        parent = "orphan" if self._parent is None else "active refernce"
        return f"<Point: x={self.x:g}, y={self.y:g}, {parent}>"

    # Additions for FakeLab

    def fake_update(self, parent: Any | None = None) -> None:
        """
        Is called from FontLab.UpdateFont()
        """
        self._parent = parent

    # Attributes

    @property
    def parent(self) -> Any:
        """
        Point's parent object
        """
        # TODO: What can be a point's parent?
        return self._parent

    @property
    def x(self) -> float:
        """
        Horizontal position of the point

        Returns:
            float: The x coordinate
        """
        return self._x

    @x.setter
    def x(self, value: float) -> None:
        self._x = float(value)

    @property
    def y(self) -> float:
        """
        Vertical position of the point

        Returns:
            float: The y coordinate
        """
        return self._y

    @y.setter
    def y(self, value: float) -> None:
        self._y = float(value)

    # Operations

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Point):
            return False
        return self.x == other.x and self.y == other.y

    # coerce?
    def __coerce__(self, other):
        """
        can be operated on Point, float value, Matrix and Rect
        """
        raise NotImplementedError

    def __add__(self, other: Point) -> Point:
        """
        Point must be second operand, both coordinates are added
        """
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        """
        Point must be second operand, both coordinates are subtracted
        """
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other: Point | float | Matrix) -> Point | float:
        """
        Second operand may be `Point`, `float` or `Matrix`. If second operand is
        `Point`, then result of scalar product is returned
        """
        if isinstance(other, Point):
            # Scalar product
            return self.x * other.x + self.y * other.y

        elif isinstance(other, Matrix):
            # Transform the point with the matrix
            p = Point(self)
            p.Transform(other)
            return p

        return Point(self.x * other, self.y * other)

    # Methods

    def Assign(self, p_or_x: Point | float, y: float | None = None) -> None:
        """
        Assigns new values to a Point
        """
        if isinstance(p_or_x, Point):
            # copy
            if y is not None:
                raise RuntimeError

            self._copy_constructor(p_or_x)
        else:
            # coordinates
            self._parent = None
            if p_or_x is not None:
                self.x = p_or_x
            if y is None:
                raise RuntimeError

            else:
                self.y = float(y)

    def Shift(self, p_or_x: Point | float, y: float | None = None) -> None:
        """
        Shifts Point on a position defined by p or x and y values
        """
        if isinstance(p_or_x, Point):
            p = p_or_x
        else:
            p = Point(p_or_x, y)
        self.x += p.x
        self.y += p.y

    def Add(self, p: Point) -> None:
        """
        Same as Shift(Point p)
        """
        self.Shift(p)

    def Sub(self, p: Point) -> None:
        """
        Subtracts p coordinates from the current Point
        """
        self.x -= p.x
        self.y -= p.y

    def Mul(self, s: float) -> None:
        """
        Multiplies Point's position to s value
        """
        self.x *= s
        self.y *= s

    def Transform(self, m: Matrix) -> None:
        """
        Applies Matrix transformation to the Point
        """
        m.fake_transform_point(self)
